"""Gymnasium and PettingZoo environments for satellite tasking problems."""

import functools
import logging
import os
from copy import deepcopy
from time import time_ns
from typing import Any, Callable, Generic, Iterable, Optional, TypeVar, Union

import numpy as np
from gymnasium import Env, spaces
from pettingzoo.utils.env import AgentID, ParallelEnv

from bsk_rl.comm import CommunicationMethod, NoCommunication
from bsk_rl.data import GlobalReward, NoReward
from bsk_rl.data.composition import ComposedReward
from bsk_rl.sats import Satellite
from bsk_rl.scene import Scenario
from bsk_rl.sim import Simulator
from bsk_rl.sim.world import WorldModel
from bsk_rl.utils import logging_config, vizard

logger = logging.getLogger(__name__)


SatObs = TypeVar("SatObs")
SatAct = TypeVar("SatAct")
MultiSatObs = tuple[SatObs, ...]
MultiSatAct = Iterable[SatAct]
SatArgRandomizer = Callable[[list[Satellite]], dict[Satellite, dict[str, Any]]]

NO_ACTION = int(2**31) - 1


def is_no_action(action):
    """Check if the action is a no-action placeholder."""
    if action is None:
        return True
    if isinstance(action, (int, np.integer)) and action == NO_ACTION:
        return True
    if isinstance(action, (np.ndarray, list, tuple)) and np.allclose(action, NO_ACTION):
        return True
    return False


def no_action_like(action):
    """Generate an action that is the same type and shape as the no-action placeholder."""
    if isinstance(action, (int, np.integer)):
        return NO_ACTION
    return action * 0 + NO_ACTION  # Aggressively try to convert while retaining type


class GeneralSatelliteTasking(Env, Generic[SatObs, SatAct]):
    def __init__(
        self,
        satellites: Union[Satellite, list[Satellite]],
        scenario: Optional[Scenario] = None,
        rewarder: Optional[Union[GlobalReward, list[GlobalReward]]] = None,
        world_type: Optional[type[WorldModel]] = None,
        world_args: Optional[dict[str, Any]] = None,
        communicator: Optional[CommunicationMethod] = None,
        sat_arg_randomizer: Optional[SatArgRandomizer] = None,
        sim_rate: float = 1.0,
        max_step_duration: float = 1e9,
        failure_penalty: float = -1.0,
        time_limit: Union[float, Callable] = float("inf"),
        terminate_on_time_limit: bool = False,
        generate_obs_retasking_only: bool = False,
        dtype: Optional[np.dtype] = None,
        log_level: Union[int, str] = logging.WARNING,
        log_dir: Optional[str] = None,
        vizard_dir: Optional[str] = None,
        vizard_settings: Optional[dict[str, Any]] = None,
        render_mode=None,
    ) -> None:
        """A `Gymnasium <https://gymnasium.farama.org>`_ environment adaptable to a wide range satellite tasking problems.

        These problems involve satellite(s) being tasked to complete tasks and maintain
        aliveness. These tasks often include rewards for data collection. The environment
        can be configured for any collection of satellites, including heterogenous
        constellations. Other configurable aspects are the scenario (e.g.
        imaging targets), data collection and recording, and intersatellite
        communication of data.

        The state space is a tuple containing the state of each satellite. Actions are
        assigned as a tuple of actions, one per satellite.

        Args:
            satellites: Satellite(s) to be simulated. See :ref:`bsk_rl.sats`.
            scenario: Environment the satellite is acting in; contains information
                about targets, etc. See :ref:`bsk_rl.scene`.
            rewarder: Handles recording and rewarding for data collection towards
                objectives. Can be a single rewarder or a tuple of multiple rewarders.
                See :ref:`bsk_rl.data`.
            communicator: Manages communication between satellites. See :ref:`bsk_rl.comm`.
            sat_arg_randomizer: For correlated randomization of satellites arguments. Should
                be a function that takes a list of satellites and returns a dictionary that
                maps satellites to dictionaries of satellite model arguments to be overridden.
            world_type: Type of Basilisk world model to be constructed.
            world_args: Arguments for :class:`~bsk_rl.sim.world.WorldModel` construction.
                Should be in the form of a dictionary with keys corresponding to the
                arguments of the constructor and values that are either the desired value
                or a function that takes no arguments and returns a randomized value.
            sim_rate: [s] Rate for model simulation.
            max_step_duration: [s] Maximum time to propagate sim at a step. If
                satellites are using variable interval actions, the actual step duration
                will be less than or equal to this value. It is preferable to set durations
                in the actions themselves.
            failure_penalty: Reward for satellite failure. Should be nonpositive.
            time_limit: [s] Time at which to truncate the simulation. Can also be a function
                that takes no arguments and returns a float. This function will be called
                every time the environment is reset to randomize the time limit.
            terminate_on_time_limit: Send terminations signal time_limit instead of just
                truncation.
            generate_obs_retasking_only: If True, only generate observations for satellites
                that require retasking. All other satellites will receive an observation of
                zeros.
            dtype: Data type for satellite observations. If None, the data type specified
                in the satellite.
            log_level: Logging level for the environment. Default is ``WARNING``.
            log_dir: Directory to write logs to in addition to the console.
            vizard_dir: Path to save Vizard visualization files. If None, no Vizard-related
                modules will be imported.
            vizard_settings: Settings for Vizard visualization. Set in ``vizIstance.settings``.
                Additionally, the key ``vizard_rate`` can be set to the rate at which Vizard updates.
                The key ``use_simple_earth`` can be set to use a lower detail Earth shader
                that may help viewing ground locations. Other settings can be found
                `in the Basilisk documentation <https://avslab.github.io/basilisk/Vizard/vizardAdvanced/vizardSettings.html#id1>`_.
            render_mode: Unused.
        """
        self.seed = None
        self._configure_logging(log_level, log_dir)
        if vizard_dir is not None:
            vizard.VIZARD_PATH = vizard_dir
        if vizard_settings is not None and vizard_dir is None:
            logger.warning(
                "Vizard settings provided but Vizard is not enabled. Ignoring settings."
            )
        self.vizard_settings = vizard_settings if vizard_settings is not None else {}

        if isinstance(satellites, Satellite):
            satellites = [satellites]
        self.satellites = deepcopy(satellites)

        self.dtype = dtype
        while True:
            for satellite in self.satellites:
                if [sat.name for sat in self.satellites].count(satellite.name) > 1:
                    for i, sat_rename in enumerate(
                        [sat for sat in self.satellites if sat.name == satellite.name]
                    ):
                        new_name = f"{sat_rename.name}_{i}"
                        logger.warning(
                            f"Renaming satellite {sat_rename.name} to {new_name}"
                        )
                        sat_rename.name = new_name

                # Update satellite observation dtypes
                if self.dtype is not None:
                    satellite.observation_builder.dtype = self.dtype

            # Check if all satellite names are unique
            sat_names = [sat.name for sat in self.satellites]
            if len(sat_names) == len(set(sat_names)):
                break

        self.simulator: Simulator

        if sat_arg_randomizer is None:
            sat_arg_randomizer = lambda sats: {}
        self.sat_arg_randomizer = sat_arg_randomizer

        if scenario is None:
            scenario = Scenario()

        if world_type is None:
            world_type = self._minimum_world_model()
        self.world_type = world_type
        if world_args is None:
            world_args = self.world_type.default_world_args()
        self.world_args_generator = self.world_type.default_world_args(**world_args)

        self.scenario = deepcopy(scenario)
        self.scenario.link_satellites(self.satellites)

        rewarder = deepcopy(rewarder)
        if rewarder is None:
            rewarder = NoReward()
        if (
            isinstance(rewarder, Iterable)
            and not type(rewarder).__name__ == "MagicMock"
        ):
            rewarder = ComposedReward(*rewarder)
        self.rewarder = rewarder
        self.rewarder.link_scenario(self.scenario)

        if communicator is None:
            communicator = NoCommunication()
        self.communicator = deepcopy(communicator)
        self.communicator.link_satellites(self.satellites)

        self.sim_rate = sim_rate
        self.max_step_duration = max_step_duration
        self.failure_penalty = failure_penalty
        if self.failure_penalty > 0:
            logger.warn("Failure penalty should be nonpositive")
        if callable(time_limit):
            self.time_limit_generator = time_limit
        else:
            self.time_limit_generator = lambda: time_limit
        self.terminate_on_time_limit = terminate_on_time_limit
        self.latest_step_duration = 0.0
        self.render_mode = render_mode
        self.generate_obs_retasking_only = generate_obs_retasking_only

    def _minimum_world_model(self) -> type[WorldModel]:
        """Determine the minimum world model required by the satellites."""
        types = set(
            sum(
                [satellite.dyn_type._requires_world() for satellite in self.satellites],
                [],
            )
        )
        if len(types) == 1:
            return list(types)[0]
        for test_type in types:
            if all([issubclass(test_type, other_type) for other_type in types]):
                return test_type

        # Else compose all types into a new class
        class MinimumEnv(*types):
            pass

        return MinimumEnv

    def get_satellite(self, name: str) -> "Satellite":
        """Get a satellite by name.

        Args:
            name: Name of the satellite to retrieve.

        Returns:
            The satellite object with the specified name.
        """
        for sat in self.satellites:
            if sat.name == name:
                return sat
        raise ValueError(f"Satellite with name '{name}' not found.")

    def _configure_logging(self, log_level, log_dir=None):
        if isinstance(log_level, str):
            log_level = log_level.upper()
        logger = logging.getLogger("bsk_rl")
        logger.setLevel(log_level)

        # Ensure each process has its own logger to avoid conflicts when printing
        # sim timestamps. Running multiple environments in the same process in
        # parallel will cause logging times to be incorrectly reported.
        warn_new_env = False
        for handler in logger.handlers:
            if handler.filters[0].proc_id == os.getpid():
                logger.handlers.remove(handler)
                warn_new_env = True

        ch = logging.StreamHandler()
        ch.setFormatter(logging_config.SimFormatter(color_output=True))
        ch.addFilter(logging_config.ContextFilter(env=self, proc_id=os.getpid()))
        logger.addHandler(ch)
        if warn_new_env:
            logger.warning(
                f"Creating logger for new env on PID={os.getpid()}. "
                "Old environments in process may now log times incorrectly."
            )

        if log_dir is not None:
            fh = logging.FileHandler(log_dir)
            fh.setFormatter(logging_config.SimFormatter(color_output=False))
            fh.addFilter(logging_config.ContextFilter(env=self, proc_id=os.getpid()))
            logger.addHandler(fh)

    def _generate_world_args(self) -> None:
        """Instantiate world_args from any randomizers in provided world_args."""
        self.world_args = {
            k: v if not callable(v) else v()
            for k, v in self.world_args_generator.items()
        }

    def _randomize_time_limit(self) -> None:
        time_limit = self.time_limit_generator()
        time_limit = np.ceil(time_limit / self.sim_rate) * self.sim_rate
        self.time_limit = time_limit

    def reset(
        self,
        seed: Optional[int] = None,
        options=None,
    ) -> tuple[MultiSatObs, dict[str, Any]]:
        """Reconstruct the simulator and reset the scenario.

        Satellite and world arguments get randomized on reset, if :class:`~bsk_rl.GeneralSatelliteTasking` ``.world_args``
        or :class:`~bsk_rl.sats.Satellite` ``.sat_args`` includes randomization functions.

        Certain classes in ``bsk_rl`` have a ``reset_pre_sim_init`` and/or ``reset_post_sim_init``
        method. These methods are respectively called before and after the new Basilisk
        :class:`~bsk_rl.sim.Simulator` is created. These allow for reset actions that
        feed into the underlying simulation and those that are dependent on the underlying
        simulation to be performed.

        Args:
            seed: Gymnasium environment seed.
            options: Unused.

        Returns:
            observation, info
        """
        # Explicitly delete the Basilisk simulation before creating a new one.
        self.delete_simulator()

        if seed is None:
            seed = time_ns() % 2**32
        logger.info(f"Resetting environment with seed={seed}")
        self.seed = seed
        super().reset(seed=self.seed)
        np.random.seed(self.seed)

        self._randomize_time_limit()

        self.scenario.reset_overwrite_previous()
        self.rewarder.reset_overwrite_previous()
        self.communicator.reset_overwrite_previous()
        for i, satellite in enumerate(self.satellites):
            satellite.reset_overwrite_previous()
            satellite.create_vizard_data(color=vizard.get_color(i))
        self.latest_step_duration = 0.0

        self._generate_world_args()
        overrides = self.sat_arg_randomizer(self.satellites)
        for satellite in self.satellites:
            sat_overrides = overrides.get(satellite, {})
            satellite.generate_sat_args(
                utc_init=self.world_args["utc_init"], **sat_overrides
            )

        self.scenario.utc_init = self.world_args["utc_init"]

        self.scenario.reset_pre_sim_init()
        self.rewarder.reset_pre_sim_init()
        self.communicator.reset_pre_sim_init()

        for satellite in self.satellites:
            self.rewarder.create_data_store(satellite)
            self.rewarder.data += satellite.data_store.data
            satellite.reset_pre_sim_init()

        self.simulator = Simulator(
            self.satellites,
            self.world_type,
            self.world_args,
            sim_rate=self.sim_rate,
            max_step_duration=self.max_step_duration,
            time_limit=self.time_limit,
        )
        self.simulator.setup_vizard(**self.vizard_settings)

        self.scenario.reset_during_sim_init()
        self.rewarder.reset_during_sim_init()
        self.communicator.reset_during_sim_init()

        self.simulator.finish_init()

        self.scenario.reset_post_sim_init()
        self.rewarder.reset_post_sim_init()
        self.communicator.reset_post_sim_init()

        for satellite in self.satellites:
            satellite.reset_post_sim_init()
            satellite.data_store.update_from_logs()

        observation = self._get_obs()
        info = self._get_info()
        logger.info("Environment reset")
        return observation, info

    def delete_simulator(self):
        """Delete Basilisk objects.

        Only the simulator contains strong references to BSK models, so deleting it
        will delete all Basilisk objects. Enable debug-level logging to verify that the
        simulator, FSW, dynamics, and world models are all deleted on reset.
        """
        try:
            del self.simulator
        except AttributeError:
            pass

    def _get_obs(self) -> MultiSatObs:
        """Compose satellite observations into a single observation.

        Returns:
            tuple: Joint observation
        """
        if self.generate_obs_retasking_only:
            return tuple(
                (
                    satellite.get_obs()
                    if satellite.requires_retasking
                    else satellite.observation_space.low * 0
                )
                for satellite in self.satellites
            )
        else:
            return tuple(satellite.get_obs() for satellite in self.satellites)

    def _get_info(self) -> dict[str, Any]:
        """Compose satellite info into a single info dict.

        Returns:
            tuple: Joint info
        """
        info: dict[str, Any] = {
            satellite.name: {"requires_retasking": satellite.requires_retasking}
            for satellite in self.satellites
        }
        info["d_ts"] = self.latest_step_duration
        return info

    def _get_reward(self):
        """Return a scalar reward for the step."""
        reward = sum(self.reward_dict.values())
        for satellite in self.satellites:
            if not satellite.is_alive(log_failure=True):
                reward += self.failure_penalty
        return reward

    def _get_terminated(self) -> bool:
        """Return the terminated flag for the step."""
        if self.terminate_on_time_limit and self._get_truncated():
            return True
        else:
            return not all(
                satellite.is_alive() and not self.rewarder.is_terminated(satellite)
                for satellite in self.satellites
            )

    def _get_truncated(self) -> bool:
        """Return the truncated flag for the step."""
        return (self.simulator.sim_time >= self.time_limit) or any(
            self.rewarder.is_truncated(satellite) for satellite in self.satellites
        )

    @property
    def action_space(self) -> spaces.Space[MultiSatAct]:
        """Compose satellite action spaces into a tuple.

        Returns:
            Joint action space
        """
        return spaces.Tuple((satellite.action_space for satellite in self.satellites))

    @property
    def observation_space(self) -> spaces.Space[MultiSatObs]:
        """Compose satellite observation spaces into a tuple.

        Note: calls ``reset()``, which can be expensive, to determine observation size.

        Returns:
            Joint observation space
        """
        try:
            self.simulator
        except AttributeError:
            logger.info("Calling env.reset() to get observation space")
            self.reset(seed=self.seed)
        return spaces.Tuple(
            [satellite.observation_space for satellite in self.satellites]
        )

    def _step(self, actions: MultiSatAct) -> None:
        logger.debug("Stepping environment with actions: %s", actions)
        if len(actions) != len(self.satellites):
            raise ValueError("There must be the same number of actions and satellites")
        for satellite, action in zip(self.satellites, actions):
            satellite.info = []  # reset satellite info log

            if not is_no_action(action):
                satellite.requires_retasking = False
                satellite.set_action(action)
            if not satellite.is_alive():
                satellite.requires_retasking = False
            else:
                if satellite.requires_retasking:
                    satellite.logger.warning(
                        f"Requires retasking but received no task."
                    )

        previous_time = self.simulator.sim_time  # should these be recorded in simulator
        self.simulator.run()
        self.latest_step_duration = self.simulator.sim_time - previous_time

        new_data = {
            satellite.name: satellite.data_store.update_from_logs()
            for satellite in self.satellites
        }
        self.reward_dict = self.rewarder.reward(new_data)

        self.communicator.communicate()

        for satellite in self.satellites:
            if satellite.requires_retasking:
                satellite.logger.info(f"Satellite {satellite.name} requires retasking")

    def step(
        self, actions: MultiSatAct
    ) -> tuple[MultiSatObs, float, bool, bool, dict[str, Any]]:
        """Propagate the simulation, update information, and get rewards.

        Args:
            actions: Joint action for satellites

        Returns:
            observation, reward, terminated, truncated, info
        """
        logger.info("=== STARTING STEP ===")
        self._step(actions)

        observation = self._get_obs()
        reward = self._get_reward()
        terminated = self._get_terminated()
        truncated = self._get_truncated()
        info = self._get_info()
        logger.info(f"Step reward: {reward}")
        if terminated or truncated:
            logger.info(f"Episode terminated: {terminated}")
            logger.info(f"Episode truncated: {truncated}")
        else:
            logger.debug(f"Episode terminated: {terminated}")
            logger.debug(f"Episode truncated: {truncated}")
        logger.debug("Step info: %s", info)
        logger.debug("Step observation: %s", observation)
        return observation, reward, terminated, truncated, info

    def render(self) -> None:  # pragma: no cover
        """No rendering implemented."""
        return None

    def close(self) -> None:
        """Try to cleanly delete everything."""
        if self.simulator is not None:
            del self.simulator


class SatelliteTasking(GeneralSatelliteTasking, Generic[SatObs, SatAct]):
    def __init__(self, satellite: Satellite, *args, **kwargs) -> None:
        """A special case of :class:`GeneralSatelliteTasking` for one satellite.

        For compatibility with standard training APIs, actions and observations are
        directly exposed for the single satellite as opposed to being wrapped in a
        tuple.

        Args:
            satellite: Satellite to be simulated.
            *args: Passed to :class:`GeneralSatelliteTasking`.
            **kwargs: Passed to :class:`GeneralSatelliteTasking`.
        """
        super().__init__(satellites=satellite, *args, **kwargs)
        if not len(self.satellites) == 1:
            raise ValueError(
                "SatelliteTasking must be initialized with a single satellite."
            )

    @property
    def action_space(self) -> spaces.Space[SatAct]:
        """Return the single satellite action space."""
        return self.satellite.action_space

    @property
    def observation_space(self) -> spaces.Box:
        """Return the single satellite observation space."""
        super().observation_space
        return self.satellite.observation_space

    @property
    def satellite(self) -> Satellite:
        """Satellite being tasked."""
        return self.satellites[0]

    def step(self, action) -> tuple[Any, float, bool, bool, dict[str, Any]]:
        """Task the satellite with a single action."""
        return super().step([action])

    def _get_obs(self) -> Any:
        return self.satellite.get_obs()

    def _get_info(self) -> dict[str, Any]:
        info = super()._get_info()
        for k, v in info[self.satellite.name].items():
            info[k] = v
        del info[self.satellite.name]
        return info


class ConstellationTasking(
    GeneralSatelliteTasking, ParallelEnv, Generic[SatObs, SatAct, AgentID]
):
    def __init__(
        self,
        *args,
        meta_agent_groupings: Optional[dict[AgentID, list[str]]] = None,
        only_retask_idle_meta_agent_members: bool = False,
        **kwargs,
    ) -> None:
        """Implements the `PettingZoo <https://pettingzoo.farama.org>`_ parallel API for the :class:`GeneralSatelliteTasking` environment.

        Args:
            *args: Passed to :class:`GeneralSatelliteTasking`.
            meta_agent_groupings: A dictionary mapping agent names to lists of satellite names.
            only_retask_idle_meta_agent_members: If True, only satellites in a meta agent
                that require retasking will receive actions. Other actions in the meta
                agent output will be ignored. This may also be useful to control in the
                training pipeline.
            **kwargs: Passed to :class:`GeneralSatelliteTasking`.
        """
        super().__init__(*args, **kwargs)

        self.only_retask_idle_meta_agent_members = only_retask_idle_meta_agent_members

        if meta_agent_groupings is None:
            meta_agent_groupings = {}

        sats_in_meta_agents = sum(meta_agent_groupings.values(), [])
        for sat in self.satellites:
            if sat.name not in sats_in_meta_agents:
                meta_agent_groupings[sat.name] = [sat.name]

        self.meta_agent_groupings: dict[AgentID, list[Satellite]] = {
            name: [self.get_satellite(member) for member in members]
            for name, members in meta_agent_groupings.items()
        }

    def _validate_meta_agent_groupings(self):
        """Validate that meta agent groupings consist of similar action spaces."""
        for name, members in self.meta_agent_groupings.items():
            if len(members) == 0:
                raise ValueError(f"Meta agent '{name}' has no members.")
            action_space_type = type(members[0].action_space)
            for member in members:
                assert isinstance(member.action_space, action_space_type), (
                    f"Meta agent '{name}' has members with different action space types."
                )
                assert isinstance(member.observation_space, spaces.Box), (
                    f"Only Box observation spaces are supported for meta agents, "
                    f"but member '{member.name}' has {type(member.observation_space)}."
                )

    def reset(
        self, seed: int | None = None, options=None
    ) -> tuple[MultiSatObs, dict[str, Any]]:
        """Reset the environment and return PettingZoo Parallel API format."""
        self.newly_dead = []
        self._agents_last_compute_time = None
        return super().reset(seed, options)

    @property
    def agents(self) -> list[AgentID]:
        """Agents currently in the environment."""
        if (
            self._agents_last_compute_time is None
            or self._agents_last_compute_time != self.simulator.sim_time
        ):
            truncated = super()._get_truncated()
            agents = [
                agent
                for agent, satellites in self.meta_agent_groupings.items()
                if all(satellite.is_alive() for satellite in satellites)
                and not truncated
            ]
            self._agents_last_compute_time = self.simulator.sim_time
            self._agents_cache = agents
            return agents
        else:
            return self._agents_cache

    @property
    def num_agents(self) -> int:
        """Number of agents currently in the environment."""
        return len(self.agents)

    @property
    def possible_agents(self) -> list[AgentID]:
        """Return the list of all possible agents."""
        return list(self.meta_agent_groupings.keys())

    @property
    def max_num_agents(self) -> int:
        """Maximum number of agents possible in the environment."""
        return len(self.possible_agents)

    @property
    def previously_dead(self) -> list[AgentID]:
        """Return the list of agents that died at least one step ago."""
        return list(set(self.possible_agents) - set(self.agents) - set(self.newly_dead))

    @property
    def observation_spaces(self) -> dict[AgentID, spaces.Box]:
        """Return the observation space for each agent."""
        super().observation_space
        self._validate_meta_agent_groupings()

        obs_spaces = {}
        for agent, satellites in self.meta_agent_groupings.items():
            if len(satellites) == 1:
                obs_spaces[agent] = satellites[0].observation_space
            else:
                obs_spaces[agent] = spaces.Box(
                    low=np.concatenate(
                        [sat.observation_space.low for sat in satellites]
                    ),
                    high=np.concatenate(
                        [sat.observation_space.high for sat in satellites]
                    ),
                    dtype=(
                        self.dtype
                        if self.dtype
                        else satellites[0].observation_space.dtype
                    ),
                )
        return obs_spaces

    @functools.lru_cache(maxsize=None)
    def observation_space(self, agent: AgentID) -> spaces.Space[SatObs]:
        """Return the observation space for a certain agent."""
        return self.observation_spaces[agent]

    @property
    def action_spaces(self) -> dict[AgentID, spaces.Space[SatAct]]:
        """Return the action space for each agent."""
        act_spaces = {}
        for agent, satellites in self.meta_agent_groupings.items():
            if len(satellites) == 1:
                act_spaces[agent] = satellites[0].action_space
            else:
                if isinstance(satellites[0].action_space, spaces.Discrete):
                    act_spaces[agent] = spaces.MultiDiscrete(
                        [sat.action_space.n for sat in satellites]
                    )
                elif isinstance(satellites[0].action_space, spaces.Box):
                    low = np.concatenate([sat.action_space.low for sat in satellites])
                    high = np.concatenate([sat.action_space.high for sat in satellites])
                    act_spaces[agent] = spaces.Box(
                        low=low,
                        high=high,
                        dtype=(
                            self.dtype
                            if self.dtype
                            else satellites[0].action_space.dtype
                        ),
                    )
        return act_spaces

    @functools.lru_cache(maxsize=None)
    def action_space(self, agent: AgentID) -> spaces.Space[SatAct]:
        """Return the action space for a certain agent."""
        return self.action_spaces[agent]

    def _requires_retasking(self, agent: AgentID) -> bool:
        """Check if the agent requires retasking."""
        return any(
            satellite.requires_retasking
            for satellite in self.meta_agent_groupings[agent]
        )

    def _get_obs(self) -> dict[AgentID, SatObs]:
        """Format the observation per the PettingZoo Parallel API."""
        obs = {}
        for agent, satellites in self.meta_agent_groupings.items():
            # Don't generate observations for agents that are dead
            if agent in self.previously_dead:
                continue

            if self.generate_obs_retasking_only and not self._requires_retasking(agent):
                agent_obs = [
                    satellite.observation_space.low * 0 for satellite in satellites
                ]
            else:
                agent_obs = [satellite.get_obs() for satellite in satellites]

            if len(agent_obs) == 1:
                obs[agent] = agent_obs[0]
            else:
                obs[agent] = np.concatenate(agent_obs)

        return obs

    def _get_reward(self) -> dict[AgentID, float]:
        """Format the reward per the PettingZoo Parallel API."""
        satellite_rewards = {
            self.get_satellite(name): reward
            for name, reward in self.reward_dict.items()
        }
        for satellite in self.satellites:
            if not satellite.is_alive():
                if satellite in satellite_rewards:
                    satellite_rewards[satellite] += self.failure_penalty
                else:
                    satellite_rewards[satellite] = self.failure_penalty

        reward = {
            agent: sum(satellite_rewards[sat] for sat in sats)
            for agent, sats in self.meta_agent_groupings.items()
        }

        reward_keys = list(reward.keys())
        for agent in reward_keys:
            if agent in self.previously_dead:
                del reward[agent]

        return reward

    def _get_terminated(self) -> dict[AgentID, bool]:
        """Format terminations per the PettingZoo Parallel API."""
        if self.terminate_on_time_limit and super()._get_truncated():
            return {
                agent: True
                for agent in self.possible_agents
                if agent not in self.previously_dead
            }
        else:
            return {
                agent: any(
                    not sat.is_alive() or self.rewarder.is_terminated(sat)
                    for sat in satellites
                )
                for agent, satellites in self.meta_agent_groupings.items()
                if agent not in self.previously_dead
            }

    def _get_truncated(self) -> dict[AgentID, bool]:
        """Format truncations per the PettingZoo Parallel API."""
        truncated = super()._get_truncated()
        return {
            agent: truncated
            or any(self.rewarder.is_truncated(sat) for sat in satellites)
            for agent, satellites in self.meta_agent_groupings.items()
            if agent not in self.previously_dead
        }

    def _get_info(self) -> dict[AgentID, dict]:
        """Format info per the PettingZoo Parallel API."""
        info_per_sat = super()._get_info()

        # Group info by agent
        info = {}
        for agent, satellites in self.meta_agent_groupings.items():
            if agent not in self.previously_dead:
                info[agent] = {
                    "requires_retasking": any(
                        info_per_sat[sat.name]["requires_retasking"]
                        for sat in satellites
                    )
                }
                if len(satellites) > 1:
                    for satellite in satellites:
                        info[agent][satellite.name] = info_per_sat[satellite.name]

        # Identify common info
        common = {
            k: v
            for k, v in info_per_sat.items()
            if k not in [sat.name for sat in self.satellites]
        }

        # Pass common info to all agents and to __common__
        for agent in info.keys():
            for k, v in common.items():
                info[agent][k] = v
        info["__common__"] = common

        return info

    def _decompose_meta_action(
        self, agent: AgentID, action: SatAct
    ) -> dict[Satellite, SatAct]:
        """Decompose a meta agent action into satellite actions."""
        sat_to_action_map = {}
        i = 0
        for satellite in self.meta_agent_groupings[agent]:
            action_len = satellite.action_space.shape
            if len(action_len) == 0:
                action_len = 1
            else:
                action_len = action_len[0]

            if isinstance(action, (list, tuple, np.ndarray)):
                if (
                    not self.only_retask_idle_meta_agent_members
                    or satellite.requires_retasking
                ):
                    if action_len == 1:
                        sat_to_action_map[satellite] = action[i]
                    else:
                        sat_to_action_map[satellite] = action[i : i + action_len]
                else:
                    sat_to_action_map[satellite] = None
                i += action_len
            else:
                sat_to_action_map[satellite] = action

        return sat_to_action_map

    def step(
        self,
        actions: dict[AgentID, SatAct],
    ) -> tuple[
        dict[AgentID, SatObs],
        dict[AgentID, float],
        dict[AgentID, bool],
        dict[AgentID, bool],
        dict[AgentID, dict],
    ]:
        """Step the environment and return PettingZoo Parallel API format."""
        logger.info("=== STARTING STEP ===")

        previous_alive = self.agents

        sat_to_action_map = {}
        for agent, action in actions.items():
            if len(self.meta_agent_groupings[agent]) > 1:
                logger.info(f"Decomposing action for meta agent {agent}")
            sat_to_action_map.update(self._decompose_meta_action(agent, action))

        action_vector = []
        for satellite in self.satellites:
            if satellite in sat_to_action_map:
                action_vector.append(sat_to_action_map[satellite])
            else:
                action_vector.append(None)
        self._step(action_vector)

        self.newly_dead = list(set(previous_alive) - set(self.agents))

        for agent in self.newly_dead:
            for satellite in self.meta_agent_groupings[agent]:
                for attr in [
                    "_timed_terminal_event_name",
                    "_image_event_name",
                ]:
                    event_name = getattr(satellite, attr, None)
                    if event_name is not None:
                        self.simulator.delete_event(event_name)

        observation = self._get_obs()
        reward = self._get_reward()
        terminated = self._get_terminated()
        truncated = self._get_truncated()
        info = self._get_info()
        nonzero_reward = {k: v for k, v in reward.items() if v != 0}
        logger.info(f"Step reward: {nonzero_reward}")
        if any(terminated.values()):
            terminated_true = [k for k, v in terminated.items() if v]
            logger.info(f"Episode terminated: {terminated_true}")
        if any(truncated.values()):
            truncated_true = [k for k, v in truncated.items() if v]
            logger.info(f"Episode truncated: {truncated_true}")
        logger.debug("Step info: %s", info)
        logger.debug("Step observation: %s", observation)
        return observation, reward, terminated, truncated, info


__all__ = []



from unittest.mock import MagicMock, call, patch

import numpy as np
import pytest
from gymnasium import spaces

from bsk_rl import act
from bsk_rl.act.actions import ActionBuilder
from bsk_rl.act.continuous_actions import ContinuousActionBuilder
from bsk_rl.act.discrete_actions import DiscreteActionBuilder


@patch.multiple(ActionBuilder, __abstractmethods__=set())
class TestActionBuilder:
    def test_init(self):
        action_spec = [MagicMock() for _ in range(3)]
        satellite = MagicMock(action_spec=action_spec)
        ab = ActionBuilder(satellite)
        for a in ab.action_spec:
            a.link_satellite.assert_called_once()

    def test_reset_post_sim_init(self):
        ab = ActionBuilder(MagicMock(action_spec=[MagicMock() for _ in range(3)]))
        ab.reset_post_sim_init()
        for a in ab.action_spec:
            a.link_simulator.assert_called_once()
            a.reset_post_sim_init.assert_called_once()


class TestDiscreteActionBuilder:
    def test_action_space(self):
        satellite = MagicMock(
            action_spec=[MagicMock(n_actions=1), MagicMock(n_actions=2)]
        )
        ab = DiscreteActionBuilder(satellite)
        assert ab.action_space == spaces.Discrete(3)

    def test_action_description(self):
        satellite = MagicMock(
            action_spec=[
                MagicMock(n_actions=1),
                MagicMock(n_actions=2),
            ]
        )
        satellite.action_spec[0].name = "foo"
        satellite.action_spec[1].name = "bar"
        ab = DiscreteActionBuilder(satellite)
        assert ab.action_description == ["foo", "bar_0", "bar_1"]

    def test_set_action(self):
        satellite = MagicMock(
            action_spec=[
                MagicMock(n_actions=1, set_action=MagicMock(return_value="foo")),
                MagicMock(n_actions=2, set_action=MagicMock(return_value="bar")),
                MagicMock(n_actions=1, set_action=MagicMock(return_value="baz")),
            ]
        )
        ab = DiscreteActionBuilder(satellite)
        ab.set_action(0)
        assert ab.action_spec[0].set_action.call_args == call(0, prev_action_key=None)
        ab.set_action(1)
        assert ab.action_spec[1].set_action.call_args == call(0, prev_action_key="foo")
        ab.set_action(2)
        assert ab.action_spec[1].set_action.call_args == call(1, prev_action_key="bar")
        ab.set_action(3)
        assert ab.action_spec[2].set_action.call_args == call(0, prev_action_key="bar")

    def test_set_action_override(self):
        satellite = MagicMock(
            action_spec=[
                MagicMock(n_actions=1, set_action_override=None),
                MagicMock(n_actions=2, set_action_override=MagicMock()),
            ]
        )
        ab = DiscreteActionBuilder(satellite)
        ab.set_action("foo")
        assert ab.action_spec[1].set_action_override.call_args == call(
            "foo", prev_action_key=None
        )


class TestDiscreteFSWAction:
    def test_set_action(self):
        fswact = act.DiscreteFSWAction("action_fsw")
        fswact.satellite = MagicMock()
        fswact.simulator = MagicMock()
        fswact.set_action(0)
        fswact.satellite.fsw.action_fsw.assert_called_once()

    def test_set_action_again(self):
        fswact = act.DiscreteFSWAction("action_fsw")
        fswact.satellite = MagicMock()
        fswact.simulator = MagicMock()
        fswact.set_action(0, prev_action_key="action_fsw")
        fswact.satellite.fsw.action_fsw.assert_not_called()

    def test_set_action_reset(self):
        fswact = act.DiscreteFSWAction("action_fsw", reset_task=True)
        fswact.satellite = MagicMock()
        fswact.simulator = MagicMock()
        fswact.set_action(0, prev_action_key="action_fsw")
        fswact.satellite.fsw.action_fsw.assert_called_once()


class TestImage:
    target = MagicMock()
    target.id = "target_1"

    def test_image(self):
        image = act.Image(n_ahead_image=10)
        image.satellite = MagicMock()
        image.satellite.parse_target_selection.return_value = self.target
        out = image.image(5, None)
        image.satellite.task_target_for_imaging.assert_called_once_with(
            self.target, max_duration=None
        )
        assert out == "target_1"

    def test_image_retask(self):
        image = act.Image(n_ahead_image=10)
        image.satellite = MagicMock()
        image.satellite.parse_target_selection.return_value = self.target
        out = image.image(5, "target_1")
        image.satellite.enable_target_window.assert_called_once_with(
            self.target, max_duration=None
        )
        assert out == "target_1"

    def test_image_max_duration(self):
        image = act.Image(n_ahead_image=10, max_duration=100.0)
        image.satellite = MagicMock()
        image.satellite.parse_target_selection.return_value = self.target
        out = image.image(5, None)
        image.satellite.task_target_for_imaging.assert_called_once_with(
            self.target, max_duration=100.0
        )
        assert out == "target_1"

    def test_set_action(self):
        image = act.Image(n_ahead_image=10)
        image.satellite = MagicMock()
        image.image = MagicMock()
        image.set_action(5)
        image.image.assert_called_once_with(5, None)

    def test_set_action_override(self):
        image = act.Image(n_ahead_image=10)
        image.satellite = MagicMock()
        image.image = MagicMock()
        image.set_action_override("image")
        image.image.assert_called_once_with("image", None)


class TestContinuousActionBuilder:
    def test_action_space(self):
        action_space = spaces.Box(low=0, high=1, shape=(3,))
        action = MagicMock(space=action_space)
        satellite = MagicMock(action_spec=[action])
        ab = ContinuousActionBuilder(satellite)
        assert ab.action_space == action_space
        assert ab._action == action

    def test_too_many_actions(self):
        satellite = MagicMock(action_spec=[MagicMock(), MagicMock()])
        with pytest.raises(AssertionError):
            ContinuousActionBuilder(satellite)

    def test_action(self):
        action = MagicMock()
        satellite = MagicMock(action_spec=[action])
        ab = ContinuousActionBuilder(satellite)
        act = MagicMock()
        ab.set_action(act)
        ab._action.set_action.assert_called_once_with(act)


class TestImpulsiveThrust:
    @pytest.mark.parametrize("requested,actual", [(0.5, 0.5), (2.0, 1.0)])
    def test_max_dv(self, requested, actual):
        impulsive_thrust = act.ImpulsiveThrust(max_dv=1.0)
        satellite = MagicMock()
        impulsive_thrust.satellite = satellite
        satellite.simulator.sim_rate = 5.0

        impulsive_thrust.set_action(np.array([requested, 0.0, 0.0, 100]))
        assert satellite.fsw.action_impulsive_thrust.call_args[0][0][0] == actual

    def test_fsw_action(self):
        impulsive_thrust = act.ImpulsiveThrust(fsw_action="some_action")
        satellite = MagicMock()
        impulsive_thrust.satellite = satellite
        satellite.simulator.sim_rate = 5.0

        impulsive_thrust.set_action(np.array([0.5, 0.0, 0.0, 100]))
        satellite.fsw.some_action.assert_called_once()

    @pytest.mark.parametrize(
        "start_time, duration_request, actual_end",
        [(10, 8, 20), (10, 1000, 110), (10, 50, 60)],
    )
    def test_duration(self, start_time, duration_request, actual_end):
        impulsive_thrust = act.ImpulsiveThrust(max_drift_duration=100)
        satellite = MagicMock()
        satellite.simulator.sim_rate = 5.0
        satellite.simulator.sim_time = start_time
        impulsive_thrust.satellite = satellite

        impulsive_thrust.set_action(np.array([0.0, 0.0, 0.0, duration_request]))

        satellite.update_timed_terminal_event.assert_called_once_with(actual_end)



from unittest.mock import MagicMock, patch

import numpy as np
import pytest
from gymnasium import spaces

from bsk_rl import ConstellationTasking, GeneralSatelliteTasking, SatelliteTasking
from bsk_rl.data.composition import ComposedReward
from bsk_rl.sats import Satellite


class TypeA:
    pass


class TypeB:
    pass


class TypeAprime(TypeA):
    pass


class TestGeneralSatelliteTasking:
    @patch(
        "bsk_rl.GeneralSatelliteTasking.__init__",
        MagicMock(return_value=None),
    )
    def test_generate_world_args(self):
        env = GeneralSatelliteTasking()
        env.unwrapped.world_args_generator = {"a": 1, "b": lambda: 2}
        env._generate_world_args()
        assert env.unwrapped.world_args == {"a": 1, "b": 2}

    @pytest.mark.parametrize(
        "classes,result",
        [
            ([[TypeA]], TypeA),
            ([[TypeA], [TypeA]], TypeA),
            ([[TypeA], [TypeAprime]], TypeAprime),
        ],
    )
    @patch(
        "bsk_rl.GeneralSatelliteTasking.__init__",
        MagicMock(return_value=None),
    )
    def test_minimum_world_model(self, classes, result):
        env = GeneralSatelliteTasking()
        env.unwrapped.satellites = [
            MagicMock(
                dyn_type=MagicMock(_requires_world=MagicMock(return_value=class_list))
            )
            for class_list in classes
        ]
        assert env._minimum_world_model() == result

    @patch(
        "bsk_rl.GeneralSatelliteTasking.__init__",
        MagicMock(return_value=None),
    )
    def test_minimum_world_model_mixed(self):
        env = GeneralSatelliteTasking()
        env.unwrapped.satellites = [
            MagicMock(
                dyn_type=MagicMock(_requires_world=MagicMock(return_value=class_list))
            )
            for class_list in [[TypeA], [TypeB]]
        ]
        model = env._minimum_world_model()
        assert issubclass(model, TypeA)
        assert issubclass(model, TypeB)

    def test_multiple_rewarders(self):
        mock_sat = MagicMock()
        mock_sat.sat_args_generator = {}
        mock_rewarder = [MagicMock(scenario=None), MagicMock(scenario=None)]
        env = GeneralSatelliteTasking(
            satellites=[mock_sat],
            world_type=MagicMock(),
            scenario=MagicMock(),
            rewarder=mock_rewarder,
        )
        assert isinstance(env.rewarder, ComposedReward)

    @patch("bsk_rl.gym.Simulator")
    def test_reset(self, mock_sim):
        mock_sat = MagicMock()
        mock_sat.sat_args_generator = {}
        mock_rewarder = MagicMock(scenario=None)
        env = GeneralSatelliteTasking(
            satellites=[mock_sat],
            world_type=MagicMock(),
            scenario=MagicMock(),
            rewarder=mock_rewarder,
        )
        mock_sat = env.satellites[0]
        mock_rewarder = env.rewarder
        mock_rewarder.data = 0
        mock_sat.data_store.data = 100
        env.unwrapped.world_args_generator = {"utc_init": "a long time ago"}
        env.communicator = MagicMock(last_communication_time=0.0)
        env.reset()
        assert mock_rewarder.data == 100
        mock_sat.generate_sat_args.assert_called_with(utc_init="a long time ago")
        mock_sim.assert_called_once()
        mock_sat.reset_pre_sim_init.assert_called_once()
        mock_rewarder.create_data_store.assert_called_once_with(mock_sat)
        env.communicator.reset_post_sim_init.assert_called_once()
        mock_sat.reset_post_sim_init.assert_called_once()

    @pytest.mark.parametrize(
        "sat_names,expected",
        [
            (["alice"], ["alice"]),
            (["alice", "alice"], ["alice_0", "alice_1"]),
            (["alice", "bob"], ["alice", "bob"]),
            (
                ["alice", "alice", "bob", "charlie", "bob", "alice"],
                ["alice_0", "alice_1", "bob_0", "charlie", "bob_1", "alice_2"],
            ),
            (["alice", "alice_0", "alice"], ["alice_0_0", "alice_0_1", "alice_1"]),
        ],
    )
    def test_name_conflict(self, sat_names, expected):
        mock_sats = [MagicMock() for _ in range(len(sat_names))]
        for sat, name in zip(mock_sats, sat_names):
            sat.name = name
        env = GeneralSatelliteTasking(
            satellites=mock_sats,
            world_type=MagicMock(),
            scenario=MagicMock(),
            rewarder=MagicMock(scenario=None),
        )
        fixed_names = [sat.name for sat in env.satellites]
        assert len(set(fixed_names)) == len(fixed_names)
        assert fixed_names == expected

    def test_get_obs(self):
        env = GeneralSatelliteTasking(
            satellites=[MagicMock(get_obs=MagicMock(return_value=i)) for i in range(3)],
            world_type=MagicMock(),
            scenario=MagicMock(),
            rewarder=MagicMock(),
        )
        assert env._get_obs() == (0, 1, 2)

    def test_get_obs_retasking_only(self):
        env = GeneralSatelliteTasking(
            satellites=[
                MagicMock(
                    get_obs=MagicMock(return_value=[i + 1]),
                    observation_space=spaces.Box(
                        -1e9, 1e9, shape=(1,), dtype=np.float32
                    ),
                    requires_retasking=(i == 1),
                )
                for i in range(3)
            ],
            world_type=MagicMock(),
            scenario=MagicMock(),
            rewarder=MagicMock(),
            generate_obs_retasking_only=True,
        )
        assert env._get_obs() == ([0], [2], [0])

    def test_get_info(self):
        mock_sats = [MagicMock(requires_retasking=True) for _ in range(3)]
        for i, sat in enumerate(mock_sats):
            sat.name = f"sat{i}"
        env = GeneralSatelliteTasking(
            satellites=mock_sats,
            world_type=MagicMock(),
            scenario=MagicMock(),
            rewarder=MagicMock(),
        )
        env.latest_step_duration = 10.0
        expected = {
            sat.name: {"requires_retasking": True} for i, sat in enumerate(mock_sats)
        }
        expected["d_ts"] = 10.0
        assert env._get_info() == expected

    def test_action_space(self):
        env = GeneralSatelliteTasking(
            satellites=[
                MagicMock(action_space=spaces.Discrete(i + 1)) for i in range(3)
            ],
            world_type=MagicMock(),
            scenario=MagicMock(),
            rewarder=MagicMock(),
        )
        assert env.action_space == spaces.Tuple(
            (spaces.Discrete(1), spaces.Discrete(2), spaces.Discrete(3))
        )

    def test_obs_space_no_sim(self):
        env = GeneralSatelliteTasking(
            satellites=[
                MagicMock(observation_space=spaces.Discrete(i + 1)) for i in range(3)
            ],
            world_type=MagicMock(),
            scenario=MagicMock(),
            rewarder=MagicMock(),
        )
        env.seed = 123
        old_seed = env.seed
        env.reset = MagicMock()
        assert env.observation_space == spaces.Tuple(
            (spaces.Discrete(1), spaces.Discrete(2), spaces.Discrete(3))
        )
        env.reset.assert_called_once_with(seed=old_seed)

    def test_obs_space_existing_sim(self):
        env = GeneralSatelliteTasking(
            satellites=[
                MagicMock(observation_space=spaces.Discrete(i + 1)) for i in range(3)
            ],
            world_type=MagicMock(),
            scenario=MagicMock(),
            rewarder=MagicMock(),
        )
        env.unwrapped.simulator = MagicMock()
        env.reset = MagicMock()
        assert env.observation_space == spaces.Tuple(
            (spaces.Discrete(1), spaces.Discrete(2), spaces.Discrete(3))
        )
        env.reset.assert_not_called()

    def test_step(self):
        mock_sats = [MagicMock() for _ in range(2)]
        env = GeneralSatelliteTasking(
            satellites=mock_sats,
            world_type=MagicMock(),
            scenario=MagicMock(),
            communicator=MagicMock(),
            rewarder=MagicMock(
                reward=MagicMock(return_value={sat.name: 12.5 for sat in mock_sats})
            ),
        )
        env._randomize_time_limit()
        mock_sats = env.satellites
        env.unwrapped.simulator = MagicMock(sim_time=101.0)
        _, reward, _, _, info = env.step((0, 10))
        mock_sats[0].set_action.assert_called_once_with(0)
        mock_sats[1].set_action.assert_called_once_with(10)
        env.unwrapped.simulator.run.assert_called_once()
        assert env.latest_step_duration == 0.0
        for sat in mock_sats:
            sat.data_store.update_from_logs.assert_called_once()
        assert reward == 25.0

    def test_step_bad_action(self):
        mock_sats = [MagicMock() for _ in range(2)]
        env = GeneralSatelliteTasking(
            satellites=mock_sats,
            world_type=MagicMock(),
            scenario=MagicMock(),
            rewarder=MagicMock(reward=MagicMock(return_value=25.0)),
        )
        env.unwrapped.simulator = MagicMock(sim_time=101.0)
        with pytest.raises(ValueError):
            env.step((0, 10, 20))
        with pytest.raises(ValueError):
            env.step((0,))

    @pytest.mark.parametrize("sat_death", [True, False])
    @pytest.mark.parametrize("timeout", [True, False])
    @pytest.mark.parametrize("terminate_on_time_limit", [True, False])
    @pytest.mark.parametrize("rewarder_terminated", [True, False])
    @pytest.mark.parametrize("rewarder_truncated", [True, False])
    def test_step_stopped(
        self,
        sat_death,
        timeout,
        terminate_on_time_limit,
        rewarder_terminated,
        rewarder_truncated,
    ):
        mock_sats = [MagicMock() for _ in range(2)]
        env = GeneralSatelliteTasking(
            satellites=mock_sats,
            world_type=MagicMock(),
            scenario=MagicMock(),
            communicator=MagicMock(),
            rewarder=MagicMock(
                reward=MagicMock(return_value={sat.name: 12.5 for sat in mock_sats}),
                is_terminated=MagicMock(return_value=rewarder_terminated),
                is_truncated=MagicMock(return_value=rewarder_truncated),
            ),
            terminate_on_time_limit=terminate_on_time_limit,
        )
        mock_sats = env.satellites
        env.unwrapped.simulator = MagicMock(sim_time=101.0)
        if timeout:
            env.time_limit = 100.0
        else:
            env.time_limit = 1000.0

        if sat_death:
            mock_sats[1].is_alive.return_value = False

        _, _, terminated, truncated, _ = env.step((0, 10))

        assert terminated == (
            sat_death
            or ((timeout or rewarder_truncated) and terminate_on_time_limit)
            or rewarder_terminated
        )
        assert truncated == timeout or rewarder_truncated

    @patch.multiple(Satellite, __abstractmethods__=set())
    def test_step_retask_needed(self, capfd):
        mock_sat = MagicMock()
        env = SatelliteTasking(
            satellite=[mock_sat],
            world_type=MagicMock(),
            scenario=MagicMock(),
            communicator=MagicMock(),
            rewarder=MagicMock(reward=MagicMock(return_value={mock_sat.name: 25.0})),
        )
        env._randomize_time_limit()
        env.unwrapped.simulator = MagicMock(sim_time=101.0)
        env.step(None)
        assert mock_sat.requires_retasking
        mock_sat.requires_retasking = True
        env.step(None)
        assert mock_sat.requires_retasking

    def test_render(self):
        pass

    def test_close(self):
        env = GeneralSatelliteTasking(
            satellites=[MagicMock()],
            world_type=MagicMock(),
            scenario=MagicMock(),
            rewarder=MagicMock(),
        )
        env.unwrapped.simulator = MagicMock()
        env.close()
        assert not hasattr(env, "simulator")


class TestSatelliteTasking:
    @patch.multiple(
        Satellite,
        __abstractmethods__=set(),
        __init__=MagicMock(return_value=None),
    )
    def test_init(self):
        mock_sat = Satellite("sat", {})
        mock_sat.name = "sat"
        env = SatelliteTasking(
            satellite=mock_sat,
            world_type=MagicMock(),
            scenario=MagicMock(),
            rewarder=MagicMock(),
        )
        assert env.unwrapped.satellite == env.unwrapped.satellites[0]

    def test_init_multisat(self):
        with pytest.raises(ValueError):
            SatelliteTasking(
                satellite=[MagicMock(), MagicMock()],
                world_type=MagicMock(),
                scenario=MagicMock(),
                rewarder=MagicMock(),
            )

    @staticmethod
    def make_env():
        mock_sat = MagicMock()
        env = SatelliteTasking(
            satellite=[mock_sat],
            world_type=MagicMock(),
            scenario=MagicMock(),
            rewarder=MagicMock(),
        )
        return env, mock_sat

    def test_action_space(self):
        env, mock_sat = self.make_env()
        assert env.action_space == env.satellite.action_space

    @patch("bsk_rl.GeneralSatelliteTasking.observation_space")
    def test_observation_space(self, obs_patch):
        env, mock_sat = self.make_env()
        env.unwrapped.simulator = MagicMock()
        assert env.observation_space == env.satellite.observation_space

    @patch("bsk_rl.GeneralSatelliteTasking.step")
    def test_step(self, step_patch):
        env, mock_sat = self.make_env()
        env.step("action")
        step_patch.assert_called_once_with(["action"])

    def test_get_obs(self):
        env, mock_sat = self.make_env()
        assert env._get_obs() == env.satellite.get_obs()


class TestConstellationTasking:
    @patch("bsk_rl.gym.Simulator")
    @patch("bsk_rl.ConstellationTasking._get_obs")
    @patch("bsk_rl.ConstellationTasking._get_info")
    def test_reset(self, mock_sim, obs_fn, info_fn):
        mock_sat_1 = MagicMock()
        mock_sat_2 = MagicMock()
        mock_sat_1.sat_args_generator = {}
        mock_sat_2.sat_args_generator = {}
        mock_sat_1.name = mock_sat_2.name = "SomeSat"
        mock_data = MagicMock(scenario=None)
        env = ConstellationTasking(
            satellites=[mock_sat_1, mock_sat_2],
            world_type=MagicMock(),
            scenario=MagicMock(),
            rewarder=mock_data,
        )
        env.unwrapped.world_args_generator = {"utc_init": "a long time ago"}
        env.communicator = MagicMock()
        obs, info = env.reset()
        obs_fn.assert_called_once()
        info_fn.assert_called_once()

    @patch(
        "bsk_rl.GeneralSatelliteTasking._get_truncated",
        MagicMock(return_value=False),
    )
    def test_agents(self):
        env = ConstellationTasking(
            satellites=[MagicMock() for i in range(3)],
            world_type=MagicMock(),
            scenario=MagicMock(),
            rewarder=MagicMock(),
        )
        env._agents_last_compute_time = None
        env.simulator = MagicMock(sim_time=0.0)
        assert env.agents == [sat.name for sat in env.unwrapped.satellites]
        assert env.num_agents == 3
        assert env.possible_agents == [sat.name for sat in env.unwrapped.satellites]
        assert env.max_num_agents == 3

    @patch(
        "bsk_rl.GeneralSatelliteTasking._get_truncated",
        MagicMock(return_value=False),
    )
    def test_get_obs(self):
        env = ConstellationTasking(
            satellites=[MagicMock(get_obs=MagicMock(return_value=i)) for i in range(3)],
            world_type=MagicMock(),
            scenario=MagicMock(),
            rewarder=MagicMock(),
        )
        env._agents_last_compute_time = None
        env.simulator = MagicMock(sim_time=0.0)
        env.newly_dead = []
        assert env._get_obs() == {
            sat.name: i for i, sat in enumerate(env.unwrapped.satellites)
        }

    @patch(
        "bsk_rl.GeneralSatelliteTasking._get_truncated",
        MagicMock(return_value=False),
    )
    def test_get_info(self):
        mock_sats = [MagicMock(requires_retasking=True) for _ in range(3)]
        for i, sat in enumerate(mock_sats):
            sat.name = f"sat{i}"
        env = ConstellationTasking(
            satellites=mock_sats,
            world_type=MagicMock(),
            scenario=MagicMock(),
            rewarder=MagicMock(),
        )
        env._agents_last_compute_time = None
        env.simulator = MagicMock(sim_time=0.0)
        env.newly_dead = []
        env.latest_step_duration = 10.0
        expected = {
            f"sat{i}": {"requires_retasking": True, "d_ts": 10.0}
            for i, sat in enumerate(mock_sats)
        }
        expected["__common__"] = {
            "d_ts": 10.0,
        }
        assert env._get_info() == expected

    def test_action_spaces(self):
        env = ConstellationTasking(
            satellites=[
                MagicMock(action_space=spaces.Discrete(i + 1)) for i in range(3)
            ],
            world_type=MagicMock(),
            scenario=MagicMock(),
            rewarder=MagicMock(),
        )
        assert env.action_spaces == {
            env.unwrapped.satellites[0].name: spaces.Discrete(1),
            env.unwrapped.satellites[1].name: spaces.Discrete(2),
            env.unwrapped.satellites[2].name: spaces.Discrete(3),
        }

    def test_obs_spaces(self):
        env = ConstellationTasking(
            satellites=[
                MagicMock(observation_space=spaces.Box(low=0, high=i + 1, shape=(1,)))
                for i in range(3)
            ],
            world_type=MagicMock(),
            scenario=MagicMock(),
            rewarder=MagicMock(),
        )
        env.unwrapped.simulator = MagicMock()
        env.reset = MagicMock()
        assert env.observation_spaces == {
            env.unwrapped.satellites[0].name: spaces.Box(low=0, high=1, shape=(1,)),
            env.unwrapped.satellites[1].name: spaces.Box(low=0, high=2, shape=(1,)),
            env.unwrapped.satellites[2].name: spaces.Box(low=0, high=3, shape=(1,)),
        }

    @patch(
        "bsk_rl.GeneralSatelliteTasking._get_truncated",
        MagicMock(return_value=False),
    )
    def test_get_reward(self):
        env = ConstellationTasking(
            satellites=[
                MagicMock(is_alive=MagicMock(return_value=False)) for i in range(3)
            ],
            world_type=MagicMock(),
            scenario=MagicMock(),
            rewarder=MagicMock(),
            failure_penalty=-20.0,
        )
        env._agents_last_compute_time = None
        env.simulator = MagicMock(sim_time=0.0)
        env.newly_dead = [sat.name for sat in env.unwrapped.satellites]
        env.reward_dict = {
            sat.name: 10.0 for i, sat in enumerate(env.unwrapped.satellites)
        }
        assert env._get_reward() == {
            sat.name: -10.0 for i, sat in enumerate(env.unwrapped.satellites)
        }

    @patch(
        "bsk_rl.GeneralSatelliteTasking._get_truncated",
        MagicMock(return_value=False),
    )
    def test_get_reward_missing_sat(self):
        env = ConstellationTasking(
            satellites=[
                MagicMock(is_alive=MagicMock(return_value=False)) for i in range(3)
            ],
            world_type=MagicMock(),
            scenario=MagicMock(),
            rewarder=MagicMock(),
            failure_penalty=-20.0,
        )
        env._agents_last_compute_time = None
        env.simulator = MagicMock(sim_time=0.0)
        env.newly_dead = [sat.name for sat in env.unwrapped.satellites]
        env.reward_dict = {env.unwrapped.satellites[0].name: 10.0}
        assert env._get_reward() == {
            env.unwrapped.satellites[0].name: -10.0,
            env.unwrapped.satellites[1].name: -20.0,
            env.unwrapped.satellites[2].name: -20.0,
        }

    @pytest.mark.parametrize("timeout", [False, True])
    @pytest.mark.parametrize("terminate_on_time_limit", [False, True])
    def test_get_terminated(self, timeout, terminate_on_time_limit):
        mock_sats = [
            MagicMock(
                is_alive=MagicMock(return_value=True if i != 0 else False),
            )
            for i in range(3)
        ]
        for i, sat in enumerate(mock_sats):
            sat.name = f"sat{i}"
        env = ConstellationTasking(
            satellites=mock_sats,
            world_type=MagicMock(),
            scenario=MagicMock(),
            rewarder=MagicMock(
                is_terminated=MagicMock(return_value=False),
                is_truncated=MagicMock(return_value=False),
            ),
            terminate_on_time_limit=terminate_on_time_limit,
            time_limit=100,
        )
        env._randomize_time_limit()
        env._agents_last_compute_time = None
        env.simulator = MagicMock(sim_time=0.0)
        env.unwrapped.simulator = MagicMock(sim_time=101 if timeout else 99)

        if not timeout or not terminate_on_time_limit:
            env.newly_dead = [sat.name for sat in env.unwrapped.satellites]
            assert env._get_terminated() == {
                env.unwrapped.satellites[0].name: True,
                env.unwrapped.satellites[1].name: False,
                env.unwrapped.satellites[2].name: False,
            }
        else:
            env.newly_dead = [sat.name for sat in env.unwrapped.satellites]
            assert env._get_terminated() == {
                env.unwrapped.satellites[0].name: True,
                env.unwrapped.satellites[1].name: True,
                env.unwrapped.satellites[2].name: True,
            }

    @pytest.mark.parametrize("time", [99, 101])
    def test_get_truncated(self, time):
        mock_sats = [MagicMock() for _ in range(3)]
        for i, sat in enumerate(mock_sats):
            sat.name = f"sat{i}"
        env = ConstellationTasking(
            satellites=mock_sats,
            world_type=MagicMock(),
            scenario=MagicMock(),
            rewarder=MagicMock(
                is_terminated=MagicMock(return_value=False),
                is_truncated=MagicMock(return_value=False),
            ),
            time_limit=100,
        )
        env._randomize_time_limit()
        env._agents_last_compute_time = None
        env.simulator = MagicMock(sim_time=0.0)
        env.unwrapped.simulator = MagicMock(sim_time=time)
        env.newly_dead = (
            [sat.name for sat in env.unwrapped.satellites] if time >= 100 else []
        )
        assert env._get_truncated() == {
            env.unwrapped.satellites[0].name: time >= 100,
            env.unwrapped.satellites[1].name: time >= 100,
            env.unwrapped.satellites[2].name: time >= 100,
        }

    def test_time_limit_generator(self):
        mock_sats = [MagicMock() for _ in range(3)]
        for i, sat in enumerate(mock_sats):
            sat.name = f"sat{i}"
        env = ConstellationTasking(
            satellites=mock_sats,
            world_type=MagicMock(),
            scenario=MagicMock(),
            rewarder=MagicMock(),
            time_limit=lambda: np.random.uniform(10, 20),
        )
        env._randomize_time_limit()
        env._agents_last_compute_time = None
        env.simulator = MagicMock(sim_time=0.0)
        env.unwrapped.simulator = MagicMock(sim_time=100.0)
        env.newly_dead = [sat.name for sat in env.unwrapped.satellites]
        assert env._get_truncated() == {
            env.unwrapped.satellites[0].name: True,
            env.unwrapped.satellites[1].name: True,
            env.unwrapped.satellites[2].name: True,
        }

    def test_close(self):
        env = ConstellationTasking(
            satellites=[MagicMock()],
            world_type=MagicMock(),
            scenario=MagicMock(),
            rewarder=MagicMock(),
        )
        env.unwrapped.simulator = MagicMock()
        env.close()
        assert not hasattr(env, "simulator")

    @patch(
        "bsk_rl.GeneralSatelliteTasking._get_truncated",
        MagicMock(return_value=False),
    )
    def test_dead(self):
        mock_sats = [MagicMock() for _ in range(3)]
        for i, sat in enumerate(mock_sats):
            sat.name = f"sat{i}"
        env = ConstellationTasking(
            satellites=mock_sats,
            world_type=MagicMock(),
            scenario=MagicMock(),
            rewarder=MagicMock(),
        )
        env._agents_last_compute_time = None
        env.simulator = MagicMock(sim_time=0.0)
        env.unwrapped.satellites[1].is_alive = MagicMock(return_value=False)
        env.unwrapped.satellites[2].is_alive = MagicMock(return_value=False)
        env.newly_dead = [env.unwrapped.satellites[2].name]
        assert env.previously_dead == [env.unwrapped.satellites[1].name]
        assert env.agents == [env.unwrapped.satellites[0].name]
        assert env.possible_agents == [sat.name for sat in env.unwrapped.satellites]

    mst = "bsk_rl.ConstellationTasking."

    @patch(
        "bsk_rl.GeneralSatelliteTasking._get_truncated",
        MagicMock(return_value=False),
    )
    @patch(mst + "_get_obs", MagicMock())
    @patch(mst + "_get_reward", MagicMock())
    @patch(mst + "_get_terminated", MagicMock())
    @patch(mst + "_get_truncated", MagicMock())
    @patch(mst + "_get_info", MagicMock())
    @patch(
        "bsk_rl.GeneralSatelliteTasking._step",
        MagicMock(),
    )
    def test_step(self):
        mock_sats = [MagicMock(is_alive=MagicMock(return_value=True)) for _ in range(3)]
        for i, sat in enumerate(mock_sats):
            sat.name = f"sat{i}"
        env = ConstellationTasking(
            satellites=mock_sats,
            world_type=MagicMock(),
            scenario=MagicMock(),
            rewarder=MagicMock(),
        )
        env._agents_last_compute_time = None
        env.simulator = MagicMock(sim_time=0.0)

        def kill_sat_2():
            env.unwrapped.satellites[2].is_alive.return_value = False

        def side_effect():
            kill_sat_2()
            env.simulator.sim_time += 1.0

        env._step.side_effect = lambda _: side_effect()
        env.unwrapped.satellites[1].is_alive.return_value = False
        env.step(
            {
                env.unwrapped.satellites[0].name: 0,
                env.unwrapped.satellites[2].name: 2,
            }
        )

        env._step.assert_called_with([0, None, 2])
        assert env.newly_dead == [env.unwrapped.satellites[2].name]



"""Communication of data between satellites."""

import logging
from abc import ABC, abstractmethod
from copy import copy
from itertools import permutations
from typing import TYPE_CHECKING

import numpy as np
from scipy.sparse.csgraph import connected_components
from scipy.special import comb

from bsk_rl.sim.dyn import LOSCommDynModel
from bsk_rl.utils.functional import Resetable

if TYPE_CHECKING:  # pragma: no cover
    from bsk_rl.sats import Satellite

logger = logging.getLogger(__name__)


class CommunicationMethod(ABC, Resetable):
    """Base class for defining data sharing between satellites."""

    def __init__(self, min_period: float = 0.0) -> None:
        """The base communication class.

        Subclasses implement a way of determining which pairs of satellites share data
        at each environment step.

        Args:
            min_period: Minimum time between evaluation of the communication method.
        """
        self.satellites: list["Satellite"]
        self.min_period = min_period

    def reset_overwrite_previous(self) -> None:
        self.last_communication_time = 0.0

    def link_satellites(self, satellites: list["Satellite"]) -> None:
        """Link the environment satellite list to the communication method.

        Args:
            satellites: List of satellites to communicate between.
        """
        self.satellites = satellites

    @abstractmethod  # pragma: no cover
    def communication_pairs(self) -> list[tuple["Satellite", "Satellite"]]:
        """List pairs of satellite that should share data, in the form (sender, recipient).

        To define a new communication type, this method must be implemented.
        """
        pass

    def communicate(self) -> None:
        """Share data between paired satellites."""
        if (
            self.satellites[0].simulator.sim_time - self.last_communication_time
            < self.min_period
        ):
            return

        communication_pairs = self.communication_pairs()

        if len(communication_pairs) > 0:
            logger.info(
                f"Communicating data in {len(communication_pairs)} direction{'s' if len(communication_pairs) > 1 else ''}."
            )

        if len(communication_pairs) == 2 * comb(len(self.satellites), 2):
            self._communicate_all()
        else:
            for sat_sender, sat_receiver in communication_pairs:
                sat_receiver.data_store.stage_communicated_data(
                    sat_sender.data_store.data
                )
            for satellite in self.satellites:
                satellite.data_store.update_with_communicated_data()

        self.last_communication_time = self.satellites[0].simulator.sim_time

    def _communicate_all(self):
        """Optimized communication between all pairs of satellites."""
        logger.info("Optimizing data communication between all pairs of satellites")

        data_type = self.satellites[0].data_store.data.__class__
        final_data = data_type()
        for satellite in self.satellites:
            final_data += satellite.data_store.data
        for satellite in self.satellites:
            satellite.data_store.data = copy(final_data)


class NoCommunication(CommunicationMethod):
    """Implements no communication between satellites."""

    def __init__(self, *args, **kwargs):
        """Implements no communication between satellites.

        This is the default communication method if no other method is specified. Satellites
        will maintain their own :class:`~bsk_rl.data.base.DataStore` and not share data with others.
        """
        super().__init__(*args, **kwargs)

    def communication_pairs(self) -> list[tuple["Satellite", "Satellite"]]:
        """Return no communication pairs."""
        return []


class FreeCommunication(CommunicationMethod):
    """Implements free communication between every satellite at every step."""

    def __init__(self, *args, **kwargs) -> None:
        """Implements free communication between every satellite at every step.."""
        super().__init__(*args, **kwargs)

    def communication_pairs(self) -> list[tuple["Satellite", "Satellite"]]:
        """Return all possible communication pairs."""
        return list(permutations(self.satellites, 2))


class LOSCommunication(CommunicationMethod):
    """Implements communication between satellites with a direct line-of-sight."""

    # TODO only communicate data from before latest LOS time

    def __init__(self, *args, **kwargs) -> None:
        """Implements communication between satellites with a direct line-of-sight.

        At the end of each step, satellites will communicate with each other if they have a
        line-of-sight between them that is not occluded by the Earth.

        Satellites must have a dynamics model that is a subclass of
        :class:`~bsk_rl.sim.dyn.LOSCommDynModel`. to use this communication method.
        """
        super().__init__(*args, **kwargs)

    def link_satellites(self, satellites: list["Satellite"]) -> None:
        """Link the environment satellite list to the communication method.

        Args:
            satellites: List of satellites to communicate between.
        """
        super().link_satellites(satellites)
        for satellite in self.satellites:
            if not issubclass(satellite.dyn_type, LOSCommDynModel):
                raise TypeError(
                    f"Satellite dynamics type {satellite.dyn_type} must be a subclass "
                    + "of LOSCommDynModel to use LOSCommunication"
                )

    def reset_post_sim_init(self) -> None:
        """Add loggers to satellites to track line-of-sight communication."""
        super().reset_post_sim_init()

        self.los_logs = {}
        for sat_1 in self.satellites:
            assert isinstance(sat_1.dynamics, LOSCommDynModel)
            for sat_2 in self.satellites:
                if sat_1 is not sat_2:
                    if sat_1 not in self.los_logs:
                        self.los_logs[sat_1] = {}

                    msg_index = sat_1.dynamics.los_comms_ids.index(sat_2.name)
                    logger = self.los_logs[sat_1][sat_2] = (
                        sat_1.dynamics.losComms.accessOutMsgs[msg_index].recorder()
                    )

                    sat_1.simulator.AddModelToTask(
                        sat_1.dynamics.task_name, logger, ModelPriority=586
                    )

    def communication_pairs(self) -> list[tuple["Satellite", "Satellite"]]:
        """Return pairs of satellites that have line-of-sight visibility."""
        pairs = []
        for sat_1, logs in self.los_logs.items():
            for sat_2, logger in logs.items():
                if any(logger.hasAccess):
                    pairs.append((sat_1, sat_2))
                    pairs.append((sat_2, sat_1))
        return pairs

    def communicate(self) -> None:
        """Clear line-of-sight communication logs once communicated."""
        super().communicate()
        for sat_1, logs in self.los_logs.items():
            for sat_2, logger in logs.items():
                logger.clear()


class MultiDegreeCommunication(CommunicationMethod):
    """Compose with another type to use multi-degree communications."""

    def __init__(self, *args, **kwargs) -> None:
        """Compose with another communication type to propagate multi-degree communication.

        If a communication method allows satellites A and B to communicate and satellites
        B and C to communicate, MultiDegreeCommunication will allow satellites A and C to
        communicate on the same step as well.
        """
        super().__init__(*args, **kwargs)

    def communication_pairs(self) -> list[tuple["Satellite", "Satellite"]]:
        """Return pairs of satellites that are connected by a path of communication through other satellites."""
        graph = np.zeros((len(self.satellites), len(self.satellites)), dtype=bool)
        for sat_1, sat_2 in super().communication_pairs():
            graph[self.satellites.index(sat_1), self.satellites.index(sat_2)] = True

        pairs = []
        n_components, labels = connected_components(graph, directed=False)
        for comp in range(n_components):
            for i_sat_1, i_sat_2 in permutations(np.where(labels == comp)[0], 2):
                pairs.append((self.satellites[i_sat_1], self.satellites[i_sat_2]))
        return pairs


class LOSMultiCommunication(MultiDegreeCommunication, LOSCommunication):
    """Multidegree line-of-sight communication.

    Composes :class:`MultiDegreeCommunication` with :class:`LOSCommunication`.
    """

    pass


class BroadcastCommunication(CommunicationMethod):
    def communication_pairs(self) -> list[tuple["Satellite", "Satellite"]]:
        """Return pairs of satellites that used broadcast action."""
        broadcasters = []
        for satellite in self.satellites:
            if any(
                hasattr(act, "broadcast_pending") and act.broadcast_pending
                for act in satellite.action_builder.action_spec
            ):
                broadcasters.append(satellite)

        access_pairs = super().communication_pairs()
        comm_pairs = [
            (sat_sender, sat_receiver)
            for sat_sender, sat_receiver in access_pairs
            if sat_sender in broadcasters
        ]

        for satellite in broadcasters:
            for act in satellite.action_builder.action_spec:
                if hasattr(act, "broadcast_pending"):
                    act.broadcast_pending = False

        return comm_pairs


__all__ = []



import gymnasium as gym
import numpy as np
from pytest import approx

from bsk_rl import act, obs, sats
from bsk_rl.sim import dyn, fsw
from bsk_rl.utils.orbital import random_orbit


class TestImpulsiveThrust:
    class ThrustSat(sats.Satellite):
        dyn_type = dyn.BasicDynamicsModel
        fsw_type = fsw.MagicOrbitalManeuverFSWModel
        observation_spec = [obs.SatProperties(dict(prop="v_BN_N"))]
        action_spec = [act.ImpulsiveThrust()]

    env = gym.make(
        "SatelliteTasking-v1",
        satellite=ThrustSat(
            "ThrusterSat",
            sat_args=dict(dv_available_init=1000.0),
        ),
        sim_rate=0.01,
        time_limit=10000.0,
        disable_env_checker=True,
    )

    def test_thrust(self):
        observation_before, _ = self.env.reset()
        observation_after, _, _, _, _ = self.env.step([100.0, 0.0, 0.0, 0.02])
        diff = observation_after - observation_before
        assert np.allclose(diff, np.array([100.0, 0.0, 0.0]), atol=0.5)

        time_before = self.env.unwrapped.simulator.sim_time
        self.env.step([0.0, 0.0, 0.0, 5.0])
        time_after = self.env.unwrapped.simulator.sim_time
        assert time_after - time_before == approx(5.0)


class TestHillImpulsiveThrust:
    class ThrustSat(sats.Satellite):
        dyn_type = dyn.BasicDynamicsModel
        fsw_type = fsw.MagicOrbitalManeuverFSWModel
        observation_spec = [obs.SatProperties(dict(prop="v_BN_N"))]
        action_spec = [act.ImpulsiveThrustHill(chief_name="ThrusterSat")]

    env = gym.make(
        "SatelliteTasking-v1",
        satellite=ThrustSat(
            "ThrusterSat",
            sat_args=dict(
                dv_available_init=1000.0,
                oe=random_orbit(i=0.0, e=0.0, Omega=0.0, omega=0.0, f=90.0),
            ),
        ),
        sim_rate=0.01,
        time_limit=10000.0,
        disable_env_checker=True,
    )

    def test_thrust(self):
        observation_before, _ = self.env.reset()
        observation_after, _, _, _, _ = self.env.step([100.0, 0.0, 0.0, 0.02])
        diff = observation_after - observation_before
        assert np.allclose(diff, np.array([0.0, 100.0, 0.0]), atol=0.5)

        time_before = self.env.unwrapped.simulator.sim_time
        self.env.step([0.0, 0.0, 0.0, 5.0])
        time_after = self.env.unwrapped.simulator.sim_time
        assert time_after - time_before == approx(5.0)



"""RSO scenarios define the geometry of a RSO.

The geometry is defined by a set of :class:`RSOPoint` objects, which give a location and
a normal vector for the point, as well as conditions for inspection (range, illumination,
etc.). Implemented geometries include:

* :class:`SphericalRSO`: Points are generated on a sphere using the Fibonacci sphere method.

This module does not consider self-shadowing effects or inspector to RSO shadowing effects.
"""

import logging
from abc import abstractmethod
from dataclasses import dataclass
from functools import cache
from typing import TYPE_CHECKING

import numpy as np

from bsk_rl.scene.scenario import Scenario
from bsk_rl.sim.dyn import RSODynModel, RSOInspectorDynModel
from bsk_rl.sim.fsw import RSOInspectorFSWModel
from bsk_rl.utils import vizard
from bsk_rl.utils.orbital import fibonacci_sphere

if TYPE_CHECKING:  # pragma: no cover
    from bsk_rl.sats import Satellite


logger = logging.getLogger(__name__)


@dataclass
class RSOPoint:
    r_PB_B: np.ndarray
    n_B: np.ndarray
    theta_max: float
    range_max: float
    theta_solar_max: float
    min_shadow_factor: float

    def __hash__(self) -> int:
        """Hash target by unique id."""
        return hash(
            (
                self.r_PB_B[0],
                self.r_PB_B[1],
                self.r_PB_B[2],
                self.n_B[0],
                self.n_B[1],
                self.n_B[2],
                self.theta_max,
                self.range_max,
                self.theta_solar_max,
                self.min_shadow_factor,
            )
        )

    @cache
    def __str__(self) -> str:
        """String representation of the RSO point."""
        return f"RSOPoint_{self.r_PB_B}"


class RSOPoints(Scenario):
    def reset_overwrite_previous(self) -> None:
        """Overwrite target list from previous episode."""
        self.rso_points = []

    def reset_pre_sim_init(self) -> None:
        """Identify RSO and Inspector satellites."""
        self.rso_points = self.generate_points()

        # Check for RSOs and inspectors
        rsos = [sat for sat in self.satellites if issubclass(sat.dyn_type, RSODynModel)]
        if len(rsos) == 0:
            logger.warning("No RSODynModel satellites found in scenario.")
            return
        assert len(rsos) == 1, "Only one RSODynModel satellite is supported."
        self.rso = rsos[0]

        self.inspectors = [
            sat
            for sat in self.satellites
            if issubclass(sat.dyn_type, RSOInspectorDynModel)
        ]
        if len(self.inspectors) == 0:
            logger.warning("No RSOInspectorDynModel satellites found in scenario.")
            return

        return super().reset_pre_sim_init()

    def reset_during_sim_init(self) -> None:
        """Add points to dynamics and fsw of RSO."""
        assert isinstance(self.rso.dynamics, RSODynModel)

        for inspector in self.inspectors:
            self.setup_inspector_camera(inspector)

        logger.debug("Adding inspection points to RSO and inspectors")
        for point in self.rso_points:
            rso_point_model = self.rso.dynamics.add_rso_point(
                point.r_PB_B,
                point.n_B,
                point.theta_max,
                point.range_max,
                point.theta_solar_max,
                point.min_shadow_factor,
            )
            # Add point to each inspector
            for inspector in self.inspectors:
                assert isinstance(inspector.dynamics, RSOInspectorDynModel)
                assert isinstance(inspector.fsw, RSOInspectorFSWModel)
                inspector.dynamics.add_rso_point(rso_point_model)

            self.visualize_rso_point(point)

        logger.debug("Targeting RSO with inspectors")
        for inspector in self.inspectors:
            inspector.fsw.set_target_rso(self.rso)

    @vizard.visualize
    def visualize_rso_point(
        self, rso_point: RSOPoint, vizSupport=None, vizInstance=None
    ):
        """Visualize target in Vizard."""
        vizSupport.addLocation(
            vizInstance,
            stationName=str(rso_point),
            parentBodyName=self.rso.name,
            r_GP_P=list(rso_point.r_PB_B),
            gHat_P=list(rso_point.n_B),
            fieldOfView=rso_point.theta_max,
            color=vizSupport.toRGBA255("gray", alpha=0.5),
            range=float(rso_point.range_max),
        )
        vizInstance.settings.showLocationCones = -1
        vizInstance.settings.showLocationCommLines = -1
        vizInstance.settings.showLocationLabels = -1

    @vizard.visualize
    def setup_inspector_camera(
        self, inspector: "Satellite", vizSupport=None, vizInstance=None
    ) -> None:
        """Visualize camera view in Vizard panel."""
        vizSupport.createStandardCamera(
            vizInstance,
            spacecraftName=inspector.name,
            setMode=0,
            bodyTarget=self.rso.name,
            setView=0,
            fieldOfView=np.radians(10),
            displayName=inspector.name,
        )

    @abstractmethod
    def generate_points(self) -> list[RSOPoint]:
        """Generate a list of RSOPoint objects based on some spacecraft geometry."""
        pass


class SphericalRSO(RSOPoints):
    def __init__(
        self,
        n_points: int = 100,
        radius: float = 1.0,
        theta_max: float = np.radians(45),
        range_max: float = -1,
        theta_solar_max: float = np.radians(60),
        min_shadow_factor: float = 0.1,
    ):
        """Generate points on a sphere using the Fibonacci sphere method.

        Args:
            n_points: Number of points to generate on the sphere.
            radius: [m] Radius of the sphere.
            theta_max: [rad] Maximum angle from the normal for inspection.
            range_max: [m] Maximum range for inspection.
            theta_solar_max: [rad] Minimum solar incidence angle for illumination.
            min_shadow_factor: Minimum shadow factor for imaging.
        """
        self.n_points = n_points
        self.radius = radius
        self.theta_max = theta_max
        self.range_max = range_max
        self.theta_solar_max = theta_solar_max
        self.min_shadow_factor = min_shadow_factor

    def generate_points(self) -> list[RSOPoint]:
        """Generate a list of RSOPoint objects on a sphere."""
        points = []

        for point in fibonacci_sphere(self.n_points):
            r_PB_B = point * self.radius
            n_B = point
            points.append(
                RSOPoint(
                    r_PB_B,
                    n_B,
                    self.theta_max,
                    self.range_max,
                    self.theta_solar_max,
                    self.min_shadow_factor,
                )
            )

        return points


__doc_title__ = "RSO Scenarios"
__all__ = ["RSOPoint", "RSOPoints", "SphericalRSO"]



"""``bsk_rl.scene`` provides scenarios, or the underlying environment in which the satellite can collect data.

Scenarios typically correspond to certain type(s) of :ref:`bsk_rl.data` systems.

For Earth observation, the following scenarios have been implemented:

* :class:`UniformTargets`: Uniformly distributed targets to be imaged by an :class:`~bsk_rl.sats.ImagingSatellite`.
* :class:`CityTargets`: Targets distributed near population centers.
* :class:`UniformNadirScanning`: Uniformly desireable data over the surface of the Earth.

For RSO Inspection tasks, the following scenario has been implemented:

* :class:`SphericalRSO`: A RSO with spherical points and radial normals.

These RSO scenarios can be used with :class:`RSOInspectionReward`.
"""

from bsk_rl.scene.rso_points import RSOPoints, SphericalRSO
from bsk_rl.scene.scenario import Scenario, UniformNadirScanning
from bsk_rl.scene.targets import CityTargets, UniformTargets

__doc_title__ = "Scenario"
__all__ = [
    "Scenario",
    "UniformTargets",
    "CityTargets",
    "UniformNadirScanning",
    "RSOPoints",
    "SphericalRSO",
]



from unittest.mock import MagicMock, patch

import pytest

from bsk_rl.sim import Simulator


@patch("Basilisk.utilities.SimulationBaseClass.SimBaseClass.__init__")
class TestSimulator:
    dyn = MagicMock()
    fsw = MagicMock()

    class MockEnv(MagicMock):
        def __init__(self, sim, rate, **kwargs):
            super().__init__(**kwargs)
            self.sim = sim
            self.rate = rate

    def mock_sim(self, **kwargs):
        Simulator.InitializeSimulation = MagicMock()
        Simulator.ConfigureStopTime = MagicMock()
        Simulator.ExecuteSimulation = MagicMock()
        Simulator.eventMap = {}

        sat = MagicMock(
            set_dynamics=MagicMock(return_value=self.dyn),
            set_fsw=MagicMock(return_value=self.fsw),
        )
        sat.name = "sat_1"
        sim = Simulator([sat], world_type=self.MockEnv, world_args={}, **kwargs)
        sim.TotalSim = MagicMock(CurrentNanos=1000000000)

        return sim

    def test_init(self, simbase_init):
        sim = self.mock_sim()
        assert sim.dynamics_list == {"sat_1": self.dyn}
        assert sim.fsw_list == {"sat_1": self.fsw}

    def test_sim_time_ns(self, simbase_init):
        sim = self.mock_sim()
        assert sim.sim_time_ns == 1000000000

    def test_sim_time(self, simbase_init):
        sim = self.mock_sim()
        assert sim.sim_time == 1.0

    def test_set_world(self, simbase_init):
        sim = self.mock_sim()
        assert sim.world.sim == sim
        assert sim.world.rate == sim.sim_rate

    def test_delete_event(self, simbase_init):
        sim = self.mock_sim()
        event = MagicMock()
        sim.eventMap = {"event": event, "other": MagicMock()}
        sim.delete_event("event")
        assert "event" not in sim.eventMap

    @pytest.mark.parametrize(
        "start_time,step_duration,time_limit,stop_time",
        [(0, 100, 50, 50), (0, 100, 200, 100), (10, 10, 50, 20)],
    )
    def test_run(self, simbase_init, start_time, step_duration, time_limit, stop_time):
        sim = self.mock_sim(max_step_duration=step_duration, time_limit=time_limit)
        sim.TotalSim.CurrentNanos = start_time * 1000000000
        sim.run()
        sim.ConfigureStopTime.assert_called_with(time_limit * 1000000000)



"""``bsk_rl.utils.functional``: General utility functions."""

import re
import warnings
from copy import deepcopy
from functools import wraps
from typing import Any, Callable, ParamSpec, TypeVar

import numpy as np


def valid_func_name(name: str) -> str:
    """Convert a string into a valid function name.

    Args:
        name: Desired function name.

    Returns:
        Sanitized function name.
    """
    # Remove all characters except for letters, digits, and underscores
    name = re.sub(r"\W+", "_", name)
    # If the name starts with a digit, add an underscore to the beginning
    if name[0].isdigit():
        name = "_" + name
    return name


def safe_dict_merge(updates: dict, base: dict) -> dict:
    """Merge a dict with another dict, warning for conflicts.

    .. code-block:: python

        >>> safe_dict_merge(dict(a=1, b=2), dict(b=2, c=3))
        {'a': 1, 'b': 2, 'c': 3}

        >>> safe_dict_merge(dict(a=1, b=4), dict(b=2, c=3))
        Warning: Conflicting values for b: overwriting 2 with 4
        {'a': 1, 'b': 4, 'c': 3}

    Args:
        updates: Dictionary to be added to base.
        base: Base dictionary to be modified.

    Returns:
        dict: Updated copy of base.
    """
    # Updates base with a copy of elements in updates
    for k, v in updates.items():
        if k in base and base[k] != v:
            warnings.warn(f"Conflicting values for {k}: overwriting {base[k]} with {v}")
    base.update(deepcopy(updates))
    return base


P = ParamSpec("P")
T = TypeVar("T")


def default_args(**defaults) -> Callable:
    """Decorate function to enumerate default arguments for collection."""

    def inner_dec(func: Callable[P, T]) -> Callable[P, T]:
        @wraps(func)
        def inner(*args: P.args, **kwargs: P.kwargs) -> T:
            return func(*args, **kwargs)

        inner.defaults = dict(**defaults)

        return inner

    return inner_dec


def collect_default_args(object: object) -> dict[str, Any]:
    """Collect all function :class:`default_args` in an object.

    Args:
        object: Object with :class:`default_args`-decorated functions.

    Returns:
        dict: Dict of keyword-value pairs of default arguments.
    """
    defaults = {}
    for name in dir(object):
        if (
            callable(getattr(object, name))
            and not name.startswith("__")
            and hasattr(getattr(object, name), "defaults")
        ):
            safe_dict_merge(getattr(object, name).defaults, defaults)
    return defaults


def vectorize_nested_dict(dictionary: dict) -> tuple[list[str], np.ndarray]:
    """Flattens a dictionary of dictionaries, arrays, and scalars into a vector."""
    keys = list(dictionary.keys())
    values = list(dictionary.values())
    for i, value in enumerate(values):
        if isinstance(value, np.ndarray):
            values[i] = value.flatten()
            keys[i] = [keys[i] + f"[{j}]" for j in range(len(value.flatten()))]
        elif isinstance(value, list):
            keys[i] = [keys[i] + f"[{j}]" for j in range(len(value))]
        elif isinstance(value, (float, int)):
            values[i] = [value]
            keys[i] = [keys[i]]
        elif isinstance(value, dict):
            prepend = keys[i]
            keys[i], values[i] = vectorize_nested_dict(value)
            keys[i] = [prepend + "." + key for key in keys[i]]

    return list(np.concatenate(keys)), np.concatenate(values)


def aliveness_checker(func: Callable[..., bool]) -> Callable[..., bool]:
    """Decorate function to evaluate when checking for satellite aliveness."""

    @wraps(func)
    def inner(*args, log_failure=False, **kwargs) -> bool:
        self = args[0]
        alive = func(*args, **kwargs)
        if not alive and log_failure:
            self.satellite.logger.warning(f"failed {func.__name__} check")
        return alive

    inner.__doc__ = (
        "*Decorated with* :class:`~bsk_rl.utils.functional.aliveness_checker`\n\n"
        + str(func.__doc__)
    )
    inner.is_aliveness_checker = True
    return inner


def check_aliveness_checkers(model: Any, log_failure=False) -> bool:
    """Evaluate all functions with @aliveness_checker in a model.

    Args:
        model: Model to search for checkers in.
        log_failure: Whether to log to the logger on checker failure.

    Returns:
        bool: Model aliveness status.
    """
    is_alive = True
    for name in dir(model):
        if (
            not name.startswith("__")
            and not is_property(model, name)
            and callable(getattr(model, name))
            and hasattr(getattr(model, name), "is_aliveness_checker")
        ):
            is_alive = is_alive and getattr(model, name)(log_failure=log_failure)
    return is_alive


def is_property(obj: Any, attr_name: str) -> bool:
    """Check if obj has an ``@property`` called ``attr_name`` without calling it."""
    cls = type(obj)
    attribute = getattr(cls, attr_name, None)
    return attribute is not None and isinstance(attribute, property)


class AbstractClassProperty:
    def __init__(self):
        """Assign a class property to act like an abstract field."""
        self.__isabstractmethod__ = True

    def __set_name__(self, owner, name):  # noqa
        self.name = name

    def __get__(self, instance, owner):  # noqa
        if instance is None:
            return self
        raise NotImplementedError(
            f"AbstractClassProperty '{self.name}' must be set in subclass"
        )


class Resetable:
    """Base class for resetable objects.

    This class is used to define objects that are reset at the beginning of each episode.
    These include :class:`~bsk_rl.scene.Scenario`, :class:`~bsk_rl.data.GlobalReward`,
    :class:`~bsk_rl.comm.CommunicationMethod`, and :class:`~bsk_rl.sats.Satellite`.

    The :class:`~bsk_rl.GeneralSatelliteTasking.reset` method takes the following steps:

    #. ``reset_overwrite_previous`` - Overwrite attributes from previous episode.
    #. ``reset_pre_sim_init`` - Reset before simulator initialization.
    #. Initialize a new Basilisk simulator.
    #. ``reset_post_sim_init`` - Reset after simulator initialization.
    """

    def reset_overwrite_previous(self) -> None:
        """Overwrite attributes from previous episode."""
        pass

    def reset_pre_sim_init(self) -> None:
        """Reset before simulator initialization."""
        pass

    def reset_during_sim_init(self) -> None:
        """Reset after simulator models have been created but before the simulator is initialized."""
        pass

    def reset_post_sim_init(self) -> None:
        """Reset after simulator initialization."""
        pass


__doc_title__ = "Functional"
__all__ = [
    "valid_func_name",
    "safe_dict_merge",
    "default_args",
    "collect_default_args",
    "vectorize_nested_dict",
    "aliveness_checker",
    "check_aliveness_checkers",
    "is_property",
    "AbstractClassProperty",
    "Resetable",
]



"""Data logging, management, and reward calculation."""

import logging
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Callable, Optional

from deprecated import deprecated

from bsk_rl.utils.functional import Resetable

if TYPE_CHECKING:  # pragma: no cover
    from bsk_rl.sats import Satellite
    from bsk_rl.scene import Scenario


logger = logging.getLogger(__name__)

LogStateType = Any


class Data(ABC):
    """Base class for units of satellite data.

    Only needs to implement the ``__add__`` method, which is used to combine two units
    of data. This is used when adding new data from actions or communication to the
    data store.
    """

    @abstractmethod  # pragma: no cover
    def __add__(self, other: "Data") -> "Data":
        """Define the combination of two units of data."""
        pass

    def __copy__(self) -> "Data":
        """Create a shallow copy of the data."""
        return self.__class__() + self


class DataStore(ABC):
    """Base class for satellite data logging."""

    data_type: type[Data]  # Define the unit of data used by the DataStore

    def __init__(
        self, satellite: "Satellite", initial_data: Optional[Data] = None
    ) -> None:
        """Base class for satellite data logging.

        One DataStore is created for each satellite in the scenario each time the
        scenario is reset. The DataStore is responsible for generating data from the
        satellite's environment and actions by comparing the current and previous-step
        state from :class:`~DataStore.get_log_state` and returning a unit of data with
        :class:`~DataStore.compare_log_states`. These two methods must be implemented
        by subclasses.

        Args:
            satellite: Satellite which data is being stored for.
            initial_data: Initial data to start the store with. Usually comes from
                :class:`~bsk_rl.data.GlobalReward.initial_data`.
        """
        self.satellite = satellite
        self.staged_data = []

        if initial_data:
            self.data = initial_data
        else:
            self.data = self.data_type()
        self.new_data = self.data_type()

    def get_log_state(self) -> LogStateType:
        """Pull information used in determining current data contribution."""
        pass

    @abstractmethod  # pragma: no cover
    def compare_log_states(
        self, old_state: LogStateType, new_state: LogStateType
    ) -> "Data":
        """Generate a unit of data based on previous step and current step logs.

        Args:
            old_state: A previous result of :class:`~DataStore.get_log_state`.
            new_state: A newer result of :class:`~DataStore.get_log_state`.

        Returns:
            Data: New data generated by the satellite.
        """
        pass

    def update_from_logs(self) -> "Data":
        """Update the data store based on collected information.

        Returns:
            New data from the previous step.
        """
        if not hasattr(self, "log_state"):
            self.log_state = self.get_log_state()
            return self.data_type()
        old_log_state = self.log_state
        self.log_state = self.get_log_state()
        new_data = self.compare_log_states(old_log_state, self.log_state)
        self.data += new_data
        self.new_data = new_data
        return new_data

    def stage_communicated_data(self, external_data: "Data") -> None:
        """Prepare data to be added from another source, but don't add it yet.

        Works with :class:`~DataStore.update_with_communicated_data` to add data from
        other satellites without erroneously propagating it through other satellites.

        Args:
            external_data: Data from another satellite to be added
        """
        self.staged_data.append(external_data)

    def update_with_communicated_data(self) -> None:
        """Update the data store from staged data."""
        for staged in self.staged_data:
            self.data += staged
        self.staged_data = []


class GlobalReward(ABC, Resetable):
    """Base class for simulation-wide data management."""

    data_store_type: type[DataStore]  # type of DataStore managed by the GlobalReward

    @classmethod
    @property
    @deprecated(reason="datastore_type is deprecated, use data_store_type instead")
    def datastore_type(cls) -> type[DataStore]:
        """:meta private: Deprecated alias for data_store_type."""
        return cls.data_store_type

    @deprecated(reason="datastore_type is deprecated, use data_store_type instead")
    def set_data_type_deprecated(self) -> None:
        """:meta private: Deprecated alias for data_store_type."""
        self.data_type = self.datastore_type.data_type
        self.data_store_type = self.datastore_type

    def __init__(self) -> None:
        """Base class for simulation-wide data management and rewarding.

        The method :class:`calculate_reward` must be overridden by subclasses. Other
        methods may be extended as necessary for housekeeping.
        """
        self.scenario: "Scenario"
        try:
            self.data_type = self.data_store_type.data_type
        except AttributeError:
            self.set_data_type_deprecated()
        self.data_store_kwargs = {}

    def link_scenario(self, scenario: "Scenario") -> None:
        """Link the data manager to the scenario.

        Args:
            scenario: The scenario that the data manager is being used with.
        """
        self.scenario = scenario

    def reset_overwrite_previous(self) -> None:
        """Overwrite attributes from previous episode."""
        self.data = self.data_type()
        self.cum_reward = {}

    def initial_data(self, satellite: "Satellite") -> "Data":
        """Furnish the :class:`~bsk_rl.data.base.DataStore` with initial data."""
        return self.data_type()

    def create_data_store(self, satellite: "Satellite") -> None:
        """Create a data store for a satellite.

        Args:
            satellite: Satellite to create a data store for.
        """
        satellite.data_store = self.data_store_type(
            satellite,
            initial_data=self.initial_data(satellite),
            **self.data_store_kwargs,
        )
        self.cum_reward[satellite.name] = 0.0

    @abstractmethod  # pragma: no cover
    def calculate_reward(self, new_data_dict: dict[str, Data]) -> dict[str, float]:
        """Calculate step reward based on all satellite data from a step.

        Returns a dictionary of rewards for each satellite based on the new data
        generated by each satellite during the previous step, in the form:

        .. code-block:: python

            {"sat-1_id": 0.23, "sat-2_id": 0.0, ...}


        Args:
            new_data_dict: A dictionary of new data generated by each satellite, in the
                form:

                .. code-block:: python

                    {"sat-1_id": data1, "sat-2_id": data2, ...}
        """
        pass

    def reward(self, new_data_dict: dict[str, Data]) -> dict[str, float]:
        """Call :class:`calculate_reward` and log cumulative reward."""
        reward = self.calculate_reward(new_data_dict)
        for satellite_id, sat_reward in reward.items():
            self.cum_reward[satellite_id] += sat_reward

        for new_data in new_data_dict.values():
            self.data += new_data

        nonzero_reward = {k: v for k, v in reward.items() if v != 0}
        logger.info(f"Total reward: {nonzero_reward}")
        return reward

    def is_truncated(self, satellite) -> bool:
        """Check if the episode should be truncated for a satellite."""
        return False

    def is_terminated(self, satellite) -> bool:
        """Check if the episode should be terminated for a satellite."""
        return False


__doc_title__ = "Base Data"
__all__ = ["GlobalReward", "DataStore", "Data"]



"""``bsk_rl.comm`` provides methods of communication for multisatellite environments.

Communication methods are used to induce collaborative behavior between satellites.
While the :class:`~bsk_rl.data.GlobalReward` acts as a global critic for the environment,
individual satellites may not have complete environmental knowledge; for example, in a
target imaging scenario, individual satellites do not know what requests have already
been fulfilled by other satellites. With communication, satellites can share data to
improve decision-making.

Communication works by sharing data between satellites, updating each other's local
knowledge of the scenario. After each environment step,
:class:`CommunicationMethod.communication_pairs` is evaluated to determine which pairs
of satellites should share data. Then, each local :class:`~bsk_rl.data.base.DataStore` is
updated with the other satellite's data.

Configuration
-------------

Communication methods can be configured by passing an instance of :class:`CommunicationMethod`
to the ``communicator`` field of the environment constructor.

.. code-block:: python

    env = ConstellationTasking(
        ...,
        communicator=LOSMultiCommunication(),
        ...
    )


Types of Communication
----------------------

* :class:`NoCommunication`: No communication between satellites.
* :class:`FreeCommunication`: Free communication between all satellites. This method is
  cheap to evaluate, and in scenarios with many satellites or tightly clustered
  satellites, it is often functionally equivalent to more complex models.
* :class:`LOSCommunication`: Line-of-sight communication between satellites. This method
  evaluates whether a direct line of sight exists between two satellites. If so, they
  can communicate.
* :class:`MultiDegreeCommunication`: This allows for "paths" of communication between
  satellites linked by some other method.
* :class:`LOSMultiCommunication`: A combination of :class:`LOSCommunication` and
  :class:`MultiDegreeCommunication` communication. This method allows for instantaneous
  communication by satellites that are obscured from each other but have a path of
  connected satellites acting as relays between them.
"""

from bsk_rl.comm.communication import (
    BroadcastCommunication,
    CommunicationMethod,
    FreeCommunication,
    LOSCommunication,
    LOSMultiCommunication,
    MultiDegreeCommunication,
    NoCommunication,
)

__doc_title__ = "Communication"
__all__ = [
    "CommunicationMethod",
    "NoCommunication",
    "FreeCommunication",
    "LOSCommunication",
    "MultiDegreeCommunication",
    "LOSMultiCommunication",
    "BroadcastCommunication",
]



from functools import partial
from unittest.mock import MagicMock, patch

import numpy as np
import pytest

from bsk_rl import sats
from bsk_rl.sim import Simulator
from bsk_rl.sim.fsw import Task


@patch.multiple(sats.Satellite, __abstractmethods__=set())
@patch("bsk_rl.sats.Satellite.observation_spec", MagicMock())
@patch("bsk_rl.sats.Satellite.action_spec", [MagicMock()])
class TestSatellite:
    sats.Satellite.dyn_type = MagicMock(
        with_defaults=MagicMock(defaults={"a": 1, "d": np.array([4, 5])})
    )
    Task.with_defaults = MagicMock(defaults={"c": 3})
    sats.Satellite.fsw_type = MagicMock(
        with_defaults=MagicMock(defaults={"b": 2}),
        some_task=Task,
    )
    sats.Satellite.logger = MagicMock()

    def test_default_sat_args(self):
        np.testing.assert_equal(
            sats.Satellite.default_sat_args(),
            {
                "a": 1,
                "b": 2,
                "c": 3,
                "d": np.array([4, 5]),
            },
        )

    @pytest.mark.parametrize(
        "overwrite,error", [({"c": 4}, False), ({"not_c": 4}, True)]
    )
    def test_default_sat_args_overwrote(self, overwrite, error):
        if not error:
            np.testing.assert_equal(
                sats.Satellite.default_sat_args(**overwrite),
                {
                    "a": 1,
                    "b": 2,
                    "c": 4,
                    "d": np.array([4, 5]),
                },
            )
        else:
            with pytest.raises(KeyError):
                sats.Satellite.default_sat_args(**overwrite)

    def test_init_default(self):
        sat = sats.Satellite(name="TestSat", sat_args=None)
        np.testing.assert_equal(
            sat.sat_args_generator, {"a": 1, "b": 2, "c": 3, "d": np.array([4, 5])}
        )

    def test_generate_sat_args(self):
        sat = sats.Satellite(
            name="TestSat",
            sat_args={"a": 4, "b": lambda: 5},
        )
        sat.generate_sat_args(a=10, d=np.array([4, 6]))
        np.testing.assert_equal(
            sat.sat_args, {"a": 10, "b": 5, "c": 3, "d": np.array([4, 6])}
        )

    # @patch("bsk_rl.env.utils.orbital.TrajectorySimulator")
    # def test_reset_pre_sim_init(self, trajsim_patch):
    #     sat = sats.Satellite(name="TestSat", sat_args=None)
    #     sat.data_store = MagicMock(is_fresh=True)
    #     sat._generate_sat_args = MagicMock()
    #     sat.sat_args = {"utc_init": 0, "rN": 0, "vN": 0, "oe": 0, "mu": 0}
    #     sat.reset_pre_sim_init()
    #     trajsim_patch.assert_called_once()
    #     sat._generate_sat_args.assert_called_once()
    #     assert sat.info == []
    #     assert sat._timed_terminal_event_name is None

    @pytest.mark.parametrize(
        "dyn_state,fsw_state,sat_past_is_alive",
        [
            (a, b, c)
            for a in [True, False]
            for b in [True, False]
            for c in [True, False]
        ],
    )
    def test_is_alive(self, dyn_state, fsw_state, sat_past_is_alive):
        sat = sats.Satellite(name="TestSat", sat_args={})
        if sat_past_is_alive:
            sat.time_of_death = None
        else:
            sat.time_of_death = 5.0
        sat.simulator = MagicMock(sim_time=10.0)
        sat._is_alive = sat_past_is_alive
        sat.dynamics = MagicMock(is_alive=MagicMock(return_value=dyn_state))
        sat.fsw = MagicMock(is_alive=MagicMock(return_value=fsw_state))
        assert sat.is_alive() == (dyn_state and fsw_state and sat_past_is_alive)
        if not sat.is_alive():
            if sat_past_is_alive:
                assert sat.time_of_death == 10.0
            else:
                assert sat.time_of_death == 5.0
        else:
            assert sat.time_of_death is None

    def test_satellite_command(self):
        sat1 = sats.Satellite(name="TestSat_1", sat_args={})
        sat2 = sats.Satellite(name="TestSat_2", sat_args={})
        self.satellites = [sat1, sat2]
        self.get_satellite = partial(Simulator.get_satellite, self)
        assert sat1 == eval(sat1._satellite_command)
        assert sat1 != eval(sat2._satellite_command)
        assert sat2 == eval(sat2._satellite_command)

    def test_update_timed_terminal_event(self):
        pass  # Probably better with integration testing

    def test_disable_timed_event(self):
        sat = sats.Satellite(name="TestSat", sat_args={})
        sat.simulator = MagicMock(eventMap={"some_event": 1})
        sat._timed_terminal_event_name = "some_event"
        sat.disable_timed_terminal_event()
        sat.simulator.delete_event.assert_called_with("some_event")

    def test_disable_timed_event_no_event(self):
        sat = sats.Satellite(name="TestSat", sat_args={})
        sat.simulator = MagicMock(eventMap={"some_event": 1})
        sat._timed_terminal_event_name = None
        sat.disable_timed_terminal_event()
        assert not sat.simulator.delete_event.called

    def test_proxy_setters(self):
        # Must be last test or others break
        mock_sim = MagicMock()
        mock_dyn = MagicMock()
        mock_fsw = MagicMock()

        sat = sats.Satellite(name="TestSat", sat_args=None)
        sat.dyn_type.return_value = mock_dyn
        sat.fsw_type.return_value = mock_fsw
        sat.generate_sat_args()
        sat.set_simulator(mock_sim)
        assert mock_dyn == sat.set_dynamics(1.0)
        assert mock_fsw == sat.set_fsw(1.0)
        # Should be proxies, not the actual object
        assert sat.simulator is not mock_sim
        assert sat.simulator == mock_sim
        assert sat.fsw is not mock_fsw
        assert sat.fsw == mock_fsw
        assert sat.dynamics is not mock_dyn
        assert sat.dynamics == mock_dyn



from unittest.mock import MagicMock

import numpy as np
import pytest

from bsk_rl.utils import functional


@pytest.mark.parametrize("input", ["valid", "123", "money$", "Hello, world!"])
def test_valid_valid_func_name(input):
    assert functional.valid_func_name(input).isidentifier()


@pytest.mark.parametrize(
    "updates,base,expected,warn",
    [
        ({"a": 1, "b": 2}, {}, {"a": 1, "b": 2}, False),
        ({"a": 1, "b": 2}, {"a": 3, "b": 4}, {"a": 1, "b": 2}, True),
        ({}, {"a": 1, "b": 2}, {"a": 1, "b": 2}, False),
        ({"b": 2}, {"a": 1, "b": 4}, {"a": 1, "b": 2}, True),
        ({"b": 2}, {"a": 1}, {"a": 1, "b": 2}, False),
        ({}, {}, {}, False),
    ],
)
def test_safe_dict_merge(updates, base, expected, warn):
    if warn:
        with pytest.warns():
            updated = functional.safe_dict_merge(updates, base)
    else:
        updated = functional.safe_dict_merge(updates, base)
    assert updated is base
    assert updated == expected


class TestDefaultArgs:
    class C1:
        @functional.default_args(a=1)
        def foo(self, a):
            self.a = a
            return self.a

    class C2:
        @functional.default_args(b=2)
        def bar(self, b):
            self.b = b
            return self.b

    class C3:
        @functional.default_args(a=3)
        def baz(self, a):
            self.a = a
            return self.a

    def test_default_args(self):
        c = self.C1()
        assert functional.collect_default_args(c) == {"a": 1}

    def test_call(self):
        c = self.C1()
        assert c.foo(**functional.collect_default_args(c)) == 1

    def test_default_args_combined(self):
        class C12(self.C1, self.C2):
            pass

        c = C12()
        assert functional.collect_default_args(c) == {"a": 1, "b": 2}

    def test_default_args_overwrite(self):
        class C13(self.C1, self.C3):
            pass

        c = C13()
        with pytest.warns():
            assert functional.collect_default_args(c) == {"a": 1}


@pytest.mark.parametrize(
    "input,outkeys,outvec",
    [
        (
            {"alpha": np.array([1]), "b": 2, "c": [3]},
            ["alpha[0]", "b", "c[0]"],
            np.array([1, 2, 3]),
        ),
        (
            {"a": {"b": 1, "charlie": 2}, "d": 3},
            ["a.b", "a.charlie", "d"],
            np.array([1, 2, 3]),
        ),
    ],
)
def test_vectorize_nested_dict(input, outkeys, outvec):
    keys, vec = functional.vectorize_nested_dict(input)
    assert np.equal(outvec, vec).all()
    assert outkeys == keys


class TestAlivenessChecker:
    class Alive(MagicMock):
        @functional.aliveness_checker
        def is_alive(self):
            return True

    class Dead(MagicMock):
        @functional.aliveness_checker
        def is_living(self):
            return False

    class Schrodinger(Alive, Dead):
        pass

    def test_alive(self):
        a = self.Alive()
        assert functional.check_aliveness_checkers(a)

    @pytest.mark.parametrize(
        "type",
        [Dead, Schrodinger],
    )
    def test_dead(self, type):
        d = type()
        d.simulator.sim_time = 0
        d.satellite.id = "SAT"
        d.satellite._is_alive
        assert functional.check_aliveness_checkers(d, log_failure=True) is False
        d.satellite.logger.warning.assert_called_with("failed is_living check")


@pytest.mark.parametrize("prop_name,expected", [("prop", True), ("not_a_prop", False)])
def test_is_property(prop_name, expected):
    class CalledException(BaseException):
        pass

    class Class:
        @property
        def prop(self):
            # should not be called when checked
            raise CalledException

    c = Class()
    assert functional.is_property(c, prop_name) == expected



"""Basilisk dynamics models are given in ``bsk_rl.sim.dyn``.

The dynamics model is the core of the satellite simulation, representing the physical
properties of the satellite and its interactions with the environment. The dynamics model
can be customized to represent different satellite configurations, actuator models, and
instrument models.

The dynamics model is selected using the ``dyn_type`` class property of the
:class:`~bsk_rl.sats.Satellite`. Certain environment elements may require specific
dynamics models, such as :class:`~bsk_rl.comm.LOSCommunication` requiring a dynamics
model that inherits from :class:`~bsk_rl.sim.dyn.LOSCommDynModel` or :class:`~bsk_rl.sats.ImagingSatellite`
requiring a dynamics model that inherits from :class:`~bsk_rl.sim.dyn.ImagingDynModel`.

Setting Parameters
------------------

Customization of the dynamics model parameters is achieved through the ``sat_args``
dictionary passed to the :class:`~bsk_rl.sats.Satellite` constructor. This dictionary is
passed on to the dynamics model setup functions, which are called each time the simulator
is reset.

Properties
----------

The dynamics model provides a number of properties for easy access to the satellite state.
These can be accessed directly from the dynamics model instance, or in the observation
via the :class:`~bsk_rl.obs.SatProperties` observation.


Aliveness Checking
------------------

Certain functions in the dynamics model are decorated with the :func:`~bsk_rl.utils.functional.aliveness_checker`
decorator. These functions are called at each step to check if the satellite is still
operational, returning true if the satellite is still alive.

"""

from deprecated import deprecated

from bsk_rl.sim.dyn.base import BasicDynamicsModel, DynamicsModel
from bsk_rl.sim.dyn.ground_imaging import (
    ContinuousImagingDynModel,
    GroundStationDynModel,
    ImagingDynModel,
)
from bsk_rl.sim.dyn.relative_motion import (
    ConjunctionDynModel,
    LOSCommDynModel,
    MaxRangeDynModel,
)
from bsk_rl.sim.dyn.rso_inspection import RSODynModel, RSOInspectorDynModel


class FullFeaturedDynModel(GroundStationDynModel, LOSCommDynModel):
    """Convenience class for a satellite with ground station and line-of-sight comms."""

    @deprecated(
        "Create your own subclass of BasicDynamicsModel instead of using FullFeaturedDynModel."
    )
    def __init__(self, *args, **kwargs) -> None:
        """Convenience class for an imaging satellite with ground stations and line-of-sight communication."""
        super().__init__(*args, **kwargs)


__doc_title__ = "Dynamics Sims"
__all__ = [
    "DynamicsModel",
    "BasicDynamicsModel",
    "LOSCommDynModel",
    "ImagingDynModel",
    "ContinuousImagingDynModel",
    "GroundStationDynModel",
    "ConjunctionDynModel",
    "MaxRangeDynModel",
    "FullFeaturedDynModel",
    "RSODynModel",
    "RSOInspectorDynModel",
]



"""Models for ground imaging dynamics."""

from typing import TYPE_CHECKING, Iterable, Optional

import numpy as np
from Basilisk.simulation import (
    groundLocation,
    partitionedStorageUnit,
    simpleInstrument,
    simplePowerSink,
    simpleStorageUnit,
    spaceToGroundTransmitter,
)
from Basilisk.utilities import orbitalMotion

from bsk_rl.sim import world
from bsk_rl.sim.dyn import BasicDynamicsModel
from bsk_rl.utils import actuator_primitives as aP
from bsk_rl.utils.functional import aliveness_checker, default_args

if TYPE_CHECKING:  # pragma: no cover
    from bsk_rl.sim.world import WorldModel


class ImagingDynModel(BasicDynamicsModel):
    """Equips the satellite with an instrument, storage unit, and transmitter."""

    def __init__(self, *args, **kwargs) -> None:
        """Equips the satellite with an instrument, storage unit, and transmitter.

        This dynamics model is used with :class:`~bsk_rl.sats.ImagingSatellite`. It
        provides the satellite with the ability to take images of a point target. To
        enable downlink, use :class:`GroundStationDynModel` and :class:`~bsk_rl.sim.world.GroundStationWorldModel`.
        """
        super().__init__(*args, **kwargs)

    @property
    def storage_level(self):
        """Storage level [bits]."""
        return self.storageUnit.storageUnitDataOutMsg.read().storageLevel

    @property
    def storage_level_fraction(self):
        """Storage level as a fraction of capacity."""
        return self.storage_level / self.storageUnit.storageCapacity

    def _setup_dynamics_objects(self, **kwargs) -> None:
        super()._setup_dynamics_objects(**kwargs)
        self.setup_instrument_power_sink(**kwargs)
        self.setup_transmitter_power_sink(**kwargs)
        self.setup_instrument(**kwargs)
        self.setup_transmitter(**kwargs)
        self.setup_storage_unit(**kwargs)
        self.setup_imaging_target(**kwargs)

    @default_args(instrumentBaudRate=8e6)
    def setup_instrument(
        self, instrumentBaudRate: float, priority: int = 895, **kwargs
    ) -> None:
        """Set up the instrument data collection model.

        Args:
            instrumentBaudRate: [bits] Data generated by an image.
            priority: Model priority.
            kwargs: Passed to other setup functions.
        """
        self.instrument = simpleInstrument.SimpleInstrument()
        self.instrument.ModelTag = "instrument" + self.satellite.name
        self.instrument.nodeBaudRate = (
            instrumentBaudRate / self.dyn_rate
        )  # makes imaging instantaneous
        self.instrument.nodeDataName = "Instrument" + self.satellite.name
        self.simulator.AddModelToTask(
            self.task_name, self.instrument, ModelPriority=priority
        )

    @default_args(
        transmitterBaudRate=-8e6, transmitterNumBuffers=100, transmitterPacketSize=None
    )
    def setup_transmitter(
        self,
        transmitterBaudRate: float,
        instrumentBaudRate: float,
        transmitterNumBuffers: int,
        transmitterPacketSize: Optional[float] = None,
        priority: int = 798,
        **kwargs,
    ) -> None:
        """Set up the transmitter model for downlinking data.

        Args:
            transmitterBaudRate: [baud] Rate of data downlink. Should be negative.
            instrumentBaudRate: [bits] Image size, used to set packet size if ``transmitterPacketSize`` is not specified.
            transmitterNumBuffers: Number of transmitter buffers
            transmitterPacketSize: [bits] Minimum data amount to be downlinked for a partition. If ``None``, set as ``instrumentBaudRate``.
            priority: Model priority.
            kwargs: Passed to other setup functions.
        """
        if transmitterBaudRate > 0:
            self.logger.warning("transmitterBaudRate should probably be negative.")
        self.transmitter = spaceToGroundTransmitter.SpaceToGroundTransmitter()
        self.transmitter.ModelTag = "transmitter" + self.satellite.name
        self.transmitter.nodeBaudRate = transmitterBaudRate  # baud
        if transmitterPacketSize is None:
            # set packet size equal to the size of a single image
            self.transmitter.packetSize = -instrumentBaudRate  # bits
        else:
            self.transmitter.packetSize = transmitterPacketSize  # bits
        self.transmitter.numBuffers = transmitterNumBuffers
        self.simulator.AddModelToTask(
            self.task_name, self.transmitter, ModelPriority=priority
        )

    @default_args(instrumentPowerDraw=-30.0)
    def setup_instrument_power_sink(
        self, instrumentPowerDraw: float, priority: int = 897, **kwargs
    ) -> None:
        """Set the instrument power sink parameters.

        The instrument draws power when in an imaging task, representing the power cost
        of operating the instrument.

        Args:
            instrumentPowerDraw: [W] Power draw when instrument is enabled.
            priority: Model priority.
            kwargs: Passed to other setup functions.
        """
        if instrumentPowerDraw > 0:
            self.logger.warning(
                "instrumentPowerDraw should probably be zero or negative."
            )
        self.instrumentPowerSink = simplePowerSink.SimplePowerSink()
        self.instrumentPowerSink.ModelTag = "insPowerSink" + self.satellite.name
        self.instrumentPowerSink.nodePowerOut = instrumentPowerDraw
        self.simulator.AddModelToTask(
            self.task_name, self.instrumentPowerSink, ModelPriority=priority
        )
        self.powerMonitor.addPowerNodeToModel(self.instrumentPowerSink.nodePowerOutMsg)

    @default_args(transmitterPowerDraw=-15.0)
    def setup_transmitter_power_sink(
        self, transmitterPowerDraw: float, priority: int = 896, **kwargs
    ) -> None:
        """Set the transmitter power sink parameters.

        The transmitter draws power when in a downlink task, representing the power cost
        of downlinking data.

        Args:
            transmitterPowerDraw: [W] Power draw when transmitter is enabled.
            priority: Model priority.
            kwargs: Passed to other setup functions.
        """
        if transmitterPowerDraw > 0:
            self.logger.warning(
                "transmitterPowerDraw should probably be zero or negative."
            )
        self.transmitterPowerSink = simplePowerSink.SimplePowerSink()
        self.transmitterPowerSink.ModelTag = "transPowerSink" + self.satellite.name
        self.transmitterPowerSink.nodePowerOut = transmitterPowerDraw
        self.simulator.AddModelToTask(
            self.task_name, self.transmitterPowerSink, ModelPriority=priority
        )
        self.powerMonitor.addPowerNodeToModel(self.transmitterPowerSink.nodePowerOutMsg)

    @default_args(
        dataStorageCapacity=20 * 8e6,
        bufferNames=None,
        storageUnitValidCheck=False,
        storageInit=0,
    )
    def setup_storage_unit(
        self,
        dataStorageCapacity: int,
        storageUnitValidCheck: bool,
        storageInit: int,
        transmitterNumBuffers: Optional[int] = None,
        bufferNames: Optional[Iterable[str]] = None,
        priority: int = 699,
        **kwargs,
    ) -> None:
        """Configure the storage unit and its buffers.

        Separate buffers can be used to track imaging of different targets. Often, the
        buffer names will be set up by satellite based on the scenario configuration.

        Args:
            dataStorageCapacity: [bits] Maximum data that can be stored.
            transmitterNumBuffers: Number of unit buffers. Not necessary if ``bufferNames``
                are given.
            bufferNames: List of buffer names to use. Named by number if ``None``.
            storageUnitValidCheck: If ``True``, enforce that the storage level is below
                the storage capacity when checking aliveness.
            storageInit: [bits] Initial storage level.
            priority: Model priority.
            kwargs: Passed to other setup functions.
        """
        self.storageUnit = partitionedStorageUnit.PartitionedStorageUnit()
        self.storageUnit.ModelTag = "storageUnit" + self.satellite.name
        self.storageUnit.storageCapacity = dataStorageCapacity  # bits
        self.storageUnit.addDataNodeToModel(self.instrument.nodeDataOutMsg)
        self.storageUnit.addDataNodeToModel(self.transmitter.nodeDataOutMsg)
        self.storageUnitValidCheck = storageUnitValidCheck
        # Add all of the targets to the data buffer
        if bufferNames is None:
            for buffer_idx in range(transmitterNumBuffers):
                self.storageUnit.addPartition(str(buffer_idx))
        else:
            if transmitterNumBuffers is not None and transmitterNumBuffers != len(
                bufferNames
            ):
                raise ValueError(
                    "transmitterNumBuffers cannot be different than len(bufferNames)."
                )
            for buffer_name in bufferNames:
                self.storageUnit.addPartition(buffer_name)

        if storageInit != 0:
            if storageInit > dataStorageCapacity or storageInit < 0:
                self.logger.warning(
                    f"Initial storage level {storageInit} incompatible with its capacity {dataStorageCapacity}."
                )
            self.storageUnit.setDataBuffer(["STORED DATA"], [int(storageInit)])

        # Add the storage unit to the transmitter
        self.transmitter.addStorageUnitToTransmitter(
            self.storageUnit.storageUnitDataOutMsg
        )

        self.simulator.AddModelToTask(
            self.task_name, self.storageUnit, ModelPriority=priority
        )

    @aliveness_checker
    def data_storage_valid(self) -> bool:
        """Check that the buffer has not run out of space.

        Only is checked if ``storageUnitValidCheck`` is ``True``; otherwise, a full storage
        unit will prevent additional data from being stored but will not cause the satellite
        to be considered dead.
        """
        storage_check = self.storageUnitValidCheck
        if storage_check:
            return self.storage_level < self.storageUnit.storageCapacity or np.isclose(
                self.storage_level, self.storageUnit.storageCapacity
            )
        else:
            return True

    @default_args(
        groundLocationPlanetRadius=orbitalMotion.REQ_EARTH * 1e3,
        imageTargetMinimumElevation=np.radians(45.0),
        imageTargetMaximumRange=-1,
    )
    def setup_imaging_target(
        self,
        groundLocationPlanetRadius: float,
        imageTargetMinimumElevation: float,
        imageTargetMaximumRange: float,
        priority: int = 2000,
        **kwargs,
    ) -> None:
        """Add a generic imaging target to dynamics.

        The target must be updated with a particular location when used.

        Args:
            groundLocationPlanetRadius: [m] Radius of ground locations from center of planet.
            imageTargetMinimumElevation: [rad] Minimum elevation angle from target to
                satellite when imaging.
            imageTargetMaximumRange: [m] Maximum range from target to satellite when
                imaging. -1 to disable.
            priority: Model priority.
            kwargs: Passed to other setup functions.
        """
        self.imagingTarget = groundLocation.GroundLocation()
        self.imagingTarget.ModelTag = "ImagingTarget"
        self.imagingTarget.planetRadius = groundLocationPlanetRadius
        self.imagingTarget.specifyLocation(0.0, 0.0, 1000.0)
        self.imagingTarget.planetInMsg.subscribeTo(
            self.world.gravFactory.spiceObject.planetStateOutMsgs[self.world.body_index]
        )
        self.imagingTarget.minimumElevation = imageTargetMinimumElevation
        self.imagingTarget.maximumRange = imageTargetMaximumRange

        self.simulator.AddModelToTask(
            self.world.world_task_name,
            self.imagingTarget,
            ModelPriority=priority,
        )
        self.imagingTarget.addSpacecraftToModel(self.scObject.scStateOutMsg)

    def reset_for_action(self) -> None:
        """Shut off power sinks unless the transmitter or instrument is being used."""
        super().reset_for_action()
        self.transmitter.dataStatus = 0
        self.transmitterPowerSink.powerStatus = 0
        self.instrumentPowerSink.powerStatus = 0


class ContinuousImagingDynModel(ImagingDynModel):
    """Equips the satellite for continuous nadir imaging."""

    def __init__(self, *args, **kwargs) -> None:
        """Equips the satellite for continuous nadir imaging.

        Equips satellite with an instrument, storage unit, and transmitter
        for continuous nadir imaging. A single data buffer is used for storage, and data
        is accumulated continuously while imaging. The imaging target is fixed at the
        center of the Earth for nadir imaging.
        """
        super().__init__(*args, **kwargs)

    @default_args(instrumentBaudRate=8e6)
    def setup_instrument(
        self, instrumentBaudRate: float, priority: int = 895, **kwargs
    ) -> None:
        """Set up the continuous instrument model.

        Args:
            instrumentBaudRate: [baud] Data generation rate step when continuously imaging.
            priority: Model priority.
            kwargs: Passed to other setup functions.
        """
        self.instrument = simpleInstrument.SimpleInstrument()
        self.instrument.ModelTag = "instrument" + self.satellite.name
        self.instrument.nodeBaudRate = instrumentBaudRate  # make imaging instantaneous
        self.instrument.nodeDataName = "Instrument" + self.satellite.name
        self.simulator.AddModelToTask(
            self.task_name, self.instrument, ModelPriority=priority
        )

    @default_args(
        dataStorageCapacity=20 * 8e6,
        storageUnitValidCheck=False,
        storageInit=0,
    )
    def setup_storage_unit(
        self,
        dataStorageCapacity: int,
        storageUnitValidCheck: bool,
        storageInit: int,
        priority: int = 699,
        **kwargs,
    ) -> None:
        """Configure the storage unit and its buffers.

        Args:
            dataStorageCapacity: [bits] Maximum data that can be stored.
            storageUnitValidCheck: If True, check that the storage level is below the
                storage capacity.
            storageInit: [bits] Initial storage level.
            priority: Model priority.
            kwargs: Passed to other setup functions.
        """
        self.storageUnit = simpleStorageUnit.SimpleStorageUnit()
        self.storageUnit.ModelTag = "storageUnit" + self.satellite.name
        self.storageUnit.storageCapacity = dataStorageCapacity  # bits
        self.storageUnit.addDataNodeToModel(self.instrument.nodeDataOutMsg)
        self.storageUnit.addDataNodeToModel(self.transmitter.nodeDataOutMsg)
        self.storageUnitValidCheck = storageUnitValidCheck
        if storageInit > dataStorageCapacity or storageInit < 0:
            self.logger.warning(
                f"Initial storage level {storageInit} incompatible with its capacity {dataStorageCapacity}."
            )
        self.storageUnit.setDataBuffer(storageInit)

        # Add the storage unit to the transmitter
        self.transmitter.addStorageUnitToTransmitter(
            self.storageUnit.storageUnitDataOutMsg
        )

        self.simulator.AddModelToTask(
            self.task_name, self.storageUnit, ModelPriority=priority
        )

    @default_args(imageTargetMaximumRange=-1)
    def setup_imaging_target(
        self,
        imageTargetMaximumRange: float = -1,
        priority: int = 2000,
        **kwargs,
    ) -> None:
        """Add a imaging target at the center of the Earth.

        Args:
            imageTargetMaximumRange: [m] Maximum range from target to satellite when
                imaging. -1 to disable.
            priority: Model priority.
            kwargs: Passed to other setup functions.
        """
        self.imagingTarget = groundLocation.GroundLocation()
        self.imagingTarget.ModelTag = "scanningTarget"
        self.imagingTarget.planetRadius = 1e-6
        self.imagingTarget.specifyLocation(0, 0, 0)
        self.imagingTarget.planetInMsg.subscribeTo(
            self.world.gravFactory.spiceObject.planetStateOutMsgs[self.world.body_index]
        )
        self.imagingTarget.minimumElevation = np.radians(-90)
        self.imagingTarget.maximumRange = imageTargetMaximumRange

        self.simulator.AddModelToTask(
            self.world.world_task_name,
            self.imagingTarget,
            ModelPriority=priority,
        )
        self.imagingTarget.addSpacecraftToModel(self.scObject.scStateOutMsg)


class GroundStationDynModel(ImagingDynModel):
    """Model that connects satellite to world ground stations."""

    def __init__(self, *args, **kwargs) -> None:
        """Model that connects satellite to world ground stations.

        This model enables the use of ground stations defined in :class:`~bsk_rl.sim.world.GroundStationWorldModel`
        for data downlink.
        """
        super().__init__(*args, **kwargs)

    @classmethod
    def _requires_world(cls) -> list[type["WorldModel"]]:
        return super()._requires_world() + [world.GroundStationWorldModel]

    def _setup_dynamics_objects(self, **kwargs) -> None:
        super()._setup_dynamics_objects(**kwargs)
        self.setup_ground_station_locations()

    def setup_ground_station_locations(self) -> None:
        """Connect the transmitter to ground stations."""
        for groundStation in self.world.groundStations:
            groundStation.addSpacecraftToModel(self.scObject.scStateOutMsg)
            self.transmitter.addAccessMsgToTransmitter(groundStation.accessOutMsgs[-1])

            if hasattr(self.satellite, "add_location_for_access_checking"):
                self.satellite.add_location_for_access_checking(
                    object=groundStation.ModelTag,
                    r_LP_P=np.array(groundStation.r_LP_P_Init).flatten(),
                    min_elev=groundStation.minimumElevation,
                    type="ground_station",
                )


__doc_title__ = "Ground Imaging"
__all__ = [
    "ImagingDynModel",
    "ContinuousImagingDynModel",
    "GroundStationDynModel",
]



import gymnasium as gym
import pytest

from bsk_rl import act, data, obs, sats
from bsk_rl.comm import (
    FreeCommunication,
    LOSCommunication,
    LOSMultiCommunication,
    NoCommunication,
)
from bsk_rl.scene import UniformTargets
from bsk_rl.sim import dyn
from bsk_rl.utils.orbital import walker_delta

oes_visible = walker_delta(
    n_spacecraft=3,  # Number of satellites
    n_planes=1,
    rel_phasing=0,
    altitude=500 * 1e3,
    inc=45,
    clustersize=3,  # Cluster all 3 satellites together
    clusterspacing=10,  # Space satellites by a true anomaly of 30 degrees
)

oes_eclipsed = walker_delta(
    n_spacecraft=3,  # Number of satellites
    n_planes=1,
    rel_phasing=0,
    altitude=500 * 1e3,
    inc=45,
    clustersize=3,  # Cluster all 3 satellites together
    clusterspacing=60,  # Space satellites by a true anomaly of 30 degrees
)


class FullFeaturedDynModel(dyn.GroundStationDynModel, dyn.LOSCommDynModel):
    pass


class FullFeaturedSatellite(sats.ImagingSatellite):
    observation_spec = [
        obs.SatProperties(dict(prop="r_BN_P", module="dynamics", norm=6e6)),
        obs.Time(),
    ]
    action_spec = [act.Image(n_ahead_image=10)]
    dyn_type = FullFeaturedDynModel


def make_communication_env(oes, comm_type):
    satellites = [
        FullFeaturedSatellite(
            "EO-1",
            sat_args=FullFeaturedSatellite.default_sat_args(
                oe=oe,
                imageAttErrorRequirement=0.05,
                imageRateErrorRequirement=0.05,
                batteryStorageCapacity=1e6,
                storedCharge_Init=1e6,
            ),
        )
        for oe in oes
    ]
    env = gym.make(
        "GeneralSatelliteTasking-v1",
        satellites=satellites,
        scenario=UniformTargets(n_targets=1000),
        rewarder=data.UniqueImageReward(),
        communicator=comm_type(),
        sim_rate=1.0,
        time_limit=5700.0,
        disable_env_checker=True,
    )
    return env


class TestNoCommunication:
    env = make_communication_env(oes_visible, NoCommunication)

    @pytest.mark.flaky(retries=3, reruns_delay=1)
    def test_comms(self):
        self.env.reset()
        for _ in range(10):
            self.env.step([5, 5, 5])
        imagesets = [
            set(sat.data_store.data.imaged) for sat in self.env.unwrapped.satellites
        ]
        assert imagesets[0] != imagesets[1]
        assert imagesets[0] != imagesets[2]
        assert imagesets[1] != imagesets[2]


class TestFreeCommunication:
    env = make_communication_env(oes_visible, FreeCommunication)

    def test_comms(self):
        self.env.reset()
        for _ in range(10):
            self.env.step([5, 5, 5])
        imagesets = [
            set(sat.data_store.data.imaged) for sat in self.env.unwrapped.satellites
        ]
        assert imagesets[0] == imagesets[1]
        assert imagesets[1] == imagesets[2]


class TestLOSCommunication:
    env = make_communication_env(oes_visible, LOSCommunication)

    def test_comms(self):
        self.env.reset()
        for _ in range(10):
            self.env.step([5, 5, 5])
        imagesets = [
            set(sat.data_store.data.imaged) for sat in self.env.unwrapped.satellites
        ]
        assert imagesets[0].issubset(imagesets[1])
        assert imagesets[2].issubset(imagesets[1])


class TestMultiDegreeCommunication:
    env = make_communication_env(oes_visible, LOSMultiCommunication)

    def test_comms(self):
        self.env.reset()
        for _ in range(10):
            self.env.step([5, 5, 5])
        imagesets = [
            set(sat.data_store.data.imaged) for sat in self.env.unwrapped.satellites
        ]
        assert imagesets[0] == imagesets[1]
        assert imagesets[1] == imagesets[2]


class TestEclipsedLOSCommunication:
    env = make_communication_env(oes_eclipsed, LOSMultiCommunication)

    def test_comms(self):
        self.env.reset()
        for _ in range(10):
            self.env.step([5, 5, 5])
        imagesets = [
            set(sat.data_store.data.imaged) for sat in self.env.unwrapped.satellites
        ]
        assert imagesets[0] != imagesets[1]
        assert imagesets[0] != imagesets[2]
        assert imagesets[1] != imagesets[2]
        assert imagesets[1] != imagesets[2]



import types
from functools import partial

import numpy as np
import pytest

from bsk_rl import ConstellationTasking, act, data, obs, sats, scene
from bsk_rl.obs.relative_observations import rso_imaged_regions
from bsk_rl.sim import dyn, fsw
from bsk_rl.utils.orbital import fibonacci_sphere, random_circular_orbit


class RSOSat(sats.Satellite):
    observation_spec = [
        obs.SatProperties(dict(prop="one", fn=lambda _: 1.0)),
    ]
    action_spec = [act.Drift(duration=1e9)]
    dyn_type = dyn.RSODynModel
    fsw_type = fsw.BasicFSWModel


class InspectorSat(sats.Satellite):
    observation_spec = [
        obs.RelativeProperties(
            dict(
                prop="rso_imaged_regions",
                fn=partial(
                    rso_imaged_regions,
                    region_centers=fibonacci_sphere(15),
                    frame="chief_hill",
                ),
            ),
            chief_name="RSO",
        ),
        obs.Eclipse(),
    ]
    action_spec = [
        act.ImpulsiveThrustHill(
            chief_name="RSO",
            max_dv=1.0,
            max_drift_duration=5700.0 * 2,
            fsw_action="action_inspect_rso",
        )
    ]
    dyn_type = types.new_class("Dyn", (dyn.MaxRangeDynModel, dyn.RSOInspectorDynModel))
    fsw_type = types.new_class(
        "FSW",
        (
            fsw.SteeringFSWModel,
            fsw.MagicOrbitalManeuverFSWModel,
            fsw.RSOInspectorFSWModel,
        ),
    )


def inspector_sat_args(**kwargs):
    return dict(
        imageAttErrorRequirement=1.0,
        imageRateErrorRequirement=None,
        instrumentBaudRate=1,
        dataStorageCapacity=1e6,
        batteryStorageCapacity=1e9,
        storedCharge_Init=1e9,
        dv_available_init=10.0,
        chief_name="RSO",
        u_max=1.0,
        **kwargs,
    )


@pytest.mark.repeat(10)
def test_inspection():
    env = ConstellationTasking(
        satellites=[
            RSOSat(
                "RSO",
                sat_args=dict(oe=random_circular_orbit(i=0, alt=500, Omega=0, f=0)),
            ),
            InspectorSat(
                "Inspector",
                sat_args=inspector_sat_args(
                    oe=random_circular_orbit(i=0, alt=500.01, Omega=0, f=0)
                ),
                obs_type=dict,
            ),
        ],
        scenario=scene.SphericalRSO(
            n_points=100,
            radius=1.0,
            theta_max=np.radians(90),  # Generous observation angle
            range_max=250,
            theta_solar_max=np.radians(90),  # Generous illumination angle
        ),
        rewarder=(data.RSOInspectionReward()),
        time_limit=60000,
        sim_rate=1.0,
        # log_level="INFO",
    )

    observation, _ = env.reset()
    in_eclipse_start = (
        observation["Inspector"]["eclipse"][0] > observation["Inspector"]["eclipse"][1]
    )

    duration = 100
    observation, reward, _, _, _ = env.step(
        dict(RSO=0, Inspector=[0.0, 0.0, 0.0, duration])
    )
    in_eclipse_end = (
        observation["Inspector"]["eclipse"][0] > observation["Inspector"]["eclipse"][1]
    )

    # Check that nothing is imaged in eclipse
    if in_eclipse_start and in_eclipse_end:
        assert sum(observation["Inspector"]["rel_props"]["rso_imaged_regions"]) == 0
    else:
        # Orbital setup means that if not in eclipse, inspector is on the illuminated side
        assert sum(observation["Inspector"]["rel_props"]["rso_imaged_regions"]) > 0

    # Check that reward was yielded correctly
    if sum(observation["Inspector"]["rel_props"]["rso_imaged_regions"] > 0):
        assert reward["Inspector"] > 0
    else:
        assert "Inspector" not in reward or reward["Inspector"] == 0
